# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
import torch
from lhotse import CutSet, MonoCut, SupervisionSegment
from lhotse.testing.dummies import dummy_recording
from torch import tensor

from nemo.collections.common.tokenizers import SentencePieceTokenizer
from nemo.collections.common.tokenizers.sentencepiece_tokenizer import create_spt_model
from nemo.collections.multimodal.speech_llm.data.lhotse_dataset import LhotseAudioQuestionAnswerDataset
from nemo.collections.multimodal.speech_llm.parts.utils.data_utils import PromptFormatterTextProcessing


@pytest.fixture
def tokenizer(capsys, tmp_path_factory):
    TOKENIZER_TRAIN_TEXT = """
    Example system message.
    Example user message.
    Example assistant message.
    TEST
    [INST]
    [/INST]
    <s>
    </s>
    <<SYS>>
    <</SYS>>
    User: Assistant:
    user model
    Instruct Output
    \n\n
    <start_of_turn> <end_of_turn>
    <|
    |>
    <|en|> <|de|> <|fr|> <|es|> <|transcribe|> <|translate|> <|pnc|> <|nopnc|> <|startoftranscript|> <|endoftext|>
    Feel free to add new tokens for your own tests!?
    But know that if you do so, you may need to update the token IDs in the existing tests!
    So, it might be a good idea to create a new tokenizer instead when adding new prompt formats.
    """
    tmpdir = tmp_path_factory.mktemp("bpe_tokenizer")
    text_path = tmpdir / "text.txt"
    text_path.write_text(TOKENIZER_TRAIN_TEXT)
    with capsys.disabled():
        create_spt_model(
            str(text_path),
            vocab_size=512,
            sample_size=-1,
            do_lower_case=False,
            output_dir=str(tmpdir),
            remove_extra_whitespaces=True,
        )
    return SentencePieceTokenizer(str(tmpdir / "tokenizer.model"))


@pytest.fixture
def cuts():
    return CutSet(
        [
            MonoCut(
                id="ex0",
                start=0,
                duration=5.0,
                channel=0,
                supervisions=[
                    SupervisionSegment(
                        id="ex0",
                        recording_id="dummy-recording-0000",
                        start=0,
                        duration=5.0,
                        text="some transcription",
                        language="en",
                    )
                ],
                recording=dummy_recording(0, duration=5.0, with_data=True),
                custom={
                    "context": "non default prompt context",
                    "answer": "some desired answer",
                    "system_prompt": "Please answer the following based on the previous speech feature.",
                },
            ),
        ]
    )


def test_speechllm_dataset(tokenizer, cuts):
    text_processor = PromptFormatterTextProcessing(tokenizer=tokenizer, prompt_format="plain")
    dataset = LhotseAudioQuestionAnswerDataset(
        text_processor=text_processor,
        default_context="do this task",
        tokens_to_generate=0,
        pad_to_max_length=True,
        max_seq_length=64,
    )

    batch = dataset[cuts]

    expected_keys = {
        "sample_ids",
        "audio_signal",
        "audio_signal_length",
        "audio_ratio",
        "metadata",
        "tokens",
        "tokens_length",
        "labels",
        "loss_mask",
        "position_ids",
        "contexts",
        "context_lengths",
        "max_length",
        "answers",
    }
    missing_keys = expected_keys - set(batch)
    unexpected_keys = set(batch) - expected_keys
    assert not missing_keys and not unexpected_keys, f"{missing_keys=} {unexpected_keys=}"

    assert batch["sample_ids"] == ["ex0"]
    assert batch["metadata"] == [{'audio_filepath': 'ex0.wav'}]
    torch.testing.assert_close(batch["audio_ratio"], tensor([1.0]))
    torch.testing.assert_close(batch["max_length"], tensor([64]))

    assert torch.is_tensor(batch["audio_signal"])
    assert torch.is_floating_point(batch["audio_signal"])
    assert batch["audio_signal"].shape == (1, 80000)
    torch.testing.assert_close(batch["audio_signal_length"], tensor([80000], dtype=torch.int32))

    # fmt: off
    expected = tensor([[  1,  78,   9,   1,  64,  80,   5,  75,  15,   6,   1,  12,  24,  14,
               23,   6,   1,  27,  14,   9,   6,  63,   6,  76,  14,  73,   2,   1,
               56, 100,  41,  14,   9,  0,   0,   0,   0,   0,   0,   0,   0,   0,
               0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
               0,   0,   0,   0,   0,   0,   0]])
    torch.testing.assert_close(batch["tokens"], expected)
    torch.testing.assert_close(batch["tokens_length"], tensor([32]))
    assert tokenizer.ids_to_text(expected[0, :33].tolist()) == "non default prompt context some transcription"

    expected = tensor([[1, 78, 9, 1, 64, 80, 5, 75, 15, 6, 1, 12, 24, 14, 23, 6, 1, 27,
             14, 9, 6, 63, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0]])
    torch.testing.assert_close(batch["contexts"], expected)
    torch.testing.assert_close(batch["context_lengths"], tensor([23]))
    assert tokenizer.ids_to_text(expected[0, :23].tolist()) == "non default prompt context"

    expected = tensor([[76, 14, 73, 2, 1, 56, 100, 41, 14, 9, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0]])
    torch.testing.assert_close(batch["answers"], expected)
    assert tokenizer.ids_to_text(expected[0, :10].tolist()) == "some transcription"

    expected = tensor([[78, 9, 1, 64, 80, 5, 75, 15, 6, 1, 12, 24, 14, 23,
             6, 1, 27, 14, 9, 6, 63, 6, 76, 14, 73, 2, 1, 56,
             100, 41, 14, 9, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0]])
    torch.testing.assert_close(batch["labels"], expected)
    assert tokenizer.ids_to_text(expected[0, :32].tolist()) == "non default prompt context some transcription"

    torch.testing.assert_close(
        batch["position_ids"],
        tensor([[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17,
                 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35,
                 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53,
                 54, 55, 56, 57, 58, 59, 60, 61, 62, 63]])
    )

    torch.testing.assert_close(
        batch["loss_mask"],
        tensor([[0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                 0., 0., 0., 0., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1., 0., 0., 0., 0.,
                 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                 0., 0., 0., 0., 0., 0., 0., 0., 0.]])
    )
    # fmt: on


@pytest.fixture
def llama_tokenizer(capsys, tmp_path_factory):
    TOKENIZER_TRAIN_TEXT = """
    a b c d e f g h i j k l m n o p q r s t u v x y z
    A B C D E F G H I J K L M N O P Q R S T U V X Y Z
    [EOG]
    Example system message.
    Example user message.
    Example assistant message.
    TEST
    [INST]
    [/INST]
    <s>
    </s>
    <<SYS>>
    <</SYS>>
    User: Assistant:
    user model
    Instruct Output
    \n\n
    <start_of_turn> <end_of_turn>
    <|
    |>
    <|en|> <|de|> <|fr|> <|es|> <|transcribe|> <|translate|> <|pnc|> <|nopnc|> <|startoftranscript|> <|endoftext|>
    Feel free to add new tokens for your own tests!?
    But know that if you do so, you may need to update the token IDs in the existing tests!
    So, it might be a good idea to create a new tokenizer instead when adding new prompt formats.
    """
    tmpdir = tmp_path_factory.mktemp("bpe_tokenizer")
    text_path = tmpdir / "text.txt"
    text_path.write_text(TOKENIZER_TRAIN_TEXT)
    with capsys.disabled():
        create_spt_model(
            str(text_path),
            vocab_size=512,
            sample_size=-1,
            do_lower_case=False,
            output_dir=str(tmpdir),
            bos=True,
            eos=True,
            user_defined_symbols=["[INST]", "[/INST]", "<<SYS>>", "<</SYS>>", "[EOG]"],
            remove_extra_whitespaces=True,
        )
    return SentencePieceTokenizer(str(tmpdir / "tokenizer.model"))


def test_speechllm_dataset_prompt_template(llama_tokenizer, cuts):
    tokenizer = llama_tokenizer
    text_processor = PromptFormatterTextProcessing(tokenizer=tokenizer, prompt_format="llama2")
    dataset = LhotseAudioQuestionAnswerDataset(
        text_processor=text_processor,
        default_context="do this task",
        tokens_to_generate=0,
        pad_to_max_length=True,
        max_seq_length=128,
    )

    batch = dataset[cuts]
    print(batch)

    expected_keys = {
        "sample_ids",
        "audio_signal",
        "audio_signal_length",
        "audio_ratio",
        "metadata",
        "tokens",
        "tokens_length",
        "labels",
        "loss_mask",
        "position_ids",
        "contexts",
        "context_lengths",
        "max_length",
        "answers",
    }
    missing_keys = expected_keys - set(batch)
    unexpected_keys = set(batch) - expected_keys
    assert not missing_keys and not unexpected_keys, f"{missing_keys=} {unexpected_keys=}"

    assert batch["sample_ids"] == ["ex0"]
    assert batch["metadata"] == [{'audio_filepath': 'ex0.wav'}]
    torch.testing.assert_close(batch["audio_ratio"], tensor([1.0]))
    torch.testing.assert_close(batch["max_length"], tensor([128]))

    assert torch.is_tensor(batch["audio_signal"])
    assert torch.is_floating_point(batch["audio_signal"])
    assert batch["audio_signal"].shape == (1, 80000)
    torch.testing.assert_close(batch["audio_signal_length"], tensor([80000], dtype=torch.int32))

    for k in ("tokens", "contexts", "answers", "labels"):
        print(f"batch['{k}']=", tokenizer.ids_to_text(batch[k][0]))
    # fmt: off
    expected = tensor([[  1,   8,   3,   8,   5,   8, 105,  18,   9,  12,  17,   9,  41,  14,
          17,  22, 125,  43,   9, 117,  19,  18,  18,  79,  48,  15,  92,  12,
          17,   9,  42,   8,  19,  14,  43,   9,  85,  21,   9, 114,  45,  19,
          86,  17,  72,  20,   9,   9,  32,  46, 117,   9, 123,  69,   9,  25,
           8,   6,   8,  93,  14,   8,  74,  88,  12,  86,  18,  13,  85,  21,
          19,  27,  13, 116,  19,  14,  13,  78,  13,   8,   4,  72,  19,  84,
           9,   8,  65, 120,  45,  19,  14,   2,   2,   2,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2]])
    torch.testing.assert_close(batch["tokens"], expected)
    torch.testing.assert_close(batch["tokens_length"], tensor([91]))
    assert tokenizer.ids_to_text(expected[0, :91].tolist()) == "[INST] <<SYS>> Please answer the following based on the previous speech feature. <</SYS>> non default prompt context [/INST] some transcription"

    expected = tensor([[  1,   8,   3,   8,   5,   8, 105,  18,   9,  12,  17,   9,  41,  14,
          17,  22, 125,  43,   9, 117,  19,  18,  18,  79,  48,  15,  92,  12,
          17,   9,  42,   8,  19,  14,  43,   9,  85,  21,   9, 114,  45,  19,
          86,  17,  72,  20,   9,   9,  32,  46, 117,   9, 123,  69,   9,  25,
           8,   6,   8,  93,  14,   8,  74,  88,  12,  86,  18,  13,  85,  21,
          19,  27,  13, 116,  19,  14,  13,  78,  13,   8,   4,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2]])
    torch.testing.assert_close(batch["contexts"], expected)
    torch.testing.assert_close(batch["context_lengths"], tensor([81]))
    assert tokenizer.ids_to_text(expected[0, :81].tolist()) == "[INST] <<SYS>> Please answer the following based on the previous speech feature. <</SYS>> non default prompt context [/INST]"

    expected = tensor([[ 72,  19,  84,   9,   8,  65, 120,  45,  19,  14,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2]])
    torch.testing.assert_close(batch["answers"], expected)
    assert tokenizer.ids_to_text(expected[0, :11].tolist()) == "some transcription"

    expected = tensor([[  8,   3,   8,   5,   8, 105,  18,   9,  12,  17,   9,  41,  14,  17,
          22, 125,  43,   9, 117,  19,  18,  18,  79,  48,  15,  92,  12,  17,
           9,  42,   8,  19,  14,  43,   9,  85,  21,   9, 114,  45,  19,  86,
          17,  72,  20,   9,   9,  32,  46, 117,   9, 123,  69,   9,  25,   8,
           6,   8,  93,  14,   8,  74,  88,  12,  86,  18,  13,  85,  21,  19,
          27,  13, 116,  19,  14,  13,  78,  13,   8,   4,  72,  19,  84,   9,
           8,  65, 120,  45,  19,  14,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,   2,
           2]])
    torch.testing.assert_close(batch["labels"], expected)
    assert tokenizer.ids_to_text(expected[0, :90].tolist()) == "[INST] <<SYS>> Please answer the following based on the previous speech feature. <</SYS>> non default prompt context [/INST] some transcription"

    torch.testing.assert_close(
        batch["position_ids"],
        tensor([[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13,
                 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27,
                 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41,
                 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55,
                 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69,
                 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83,
                 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95, 96, 97,
                 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111,
                 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125,
                 126, 127]])
    )

    torch.testing.assert_close(
        batch["loss_mask"],
        tensor([[0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                 0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.,
                 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
                 0.]])
    )
    # fmt: on


def test_speechllm_dataset_tokens_to_generate_increases_seq_len(llama_tokenizer, cuts):
    tokenizer = llama_tokenizer
    text_processor = PromptFormatterTextProcessing(tokenizer=tokenizer, prompt_format="llama2")

    dataset = LhotseAudioQuestionAnswerDataset(
        text_processor=text_processor,
        default_context="do this task",
        tokens_to_generate=0,
        pad_to_max_length=False,
        max_seq_length=512,
    )
    batch = dataset[cuts]
    assert batch["tokens"].shape == (1, 91)
    assert batch["labels"].shape == (1, 91)
    assert batch["contexts"].shape == (1, 81)  # was 92 before padding optimization
    assert batch["answers"].shape == (1, 11)  # was 92 before padding optimization
    assert batch["position_ids"].shape == (1, 92)

    dataset = LhotseAudioQuestionAnswerDataset(
        text_processor=text_processor,
        default_context="do this task",
        tokens_to_generate=256,
        pad_to_max_length=False,
        max_seq_length=512,
    )
    batch = dataset[cuts]
    assert batch["tokens"].shape == (1, 91)
    assert batch["labels"].shape == (1, 91)
    assert batch["contexts"].shape == (1, 337)
    assert batch["answers"].shape == (1, 11)
    assert batch["position_ids"].shape == (1, 92)



# Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# Copyright (c) 2024 Arc Institute. All rights reserved.
# Copyright (c) 2024 Michael Poli. All rights reserved.
# Copyright (c) 2024 Stanford University. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from functools import partial
from typing import Literal

import torch

# CP related utils
import torch.distributed as dist
import torch.nn as nn
import torch.nn.functional as F
from megatron.core.parallel_state import (
    get_context_parallel_group,
    get_context_parallel_rank,
    get_context_parallel_world_size,
    get_tensor_model_parallel_rank,
    get_tensor_model_parallel_world_size,
)
from megatron.core.tensor_parallel import get_cuda_rng_tracker
from megatron.core.transformer.transformer_config import TransformerConfig
from megatron.core.transformer.utils import make_sharded_tensors_for_checkpoint, sharded_state_dict_default
from torch.autograd.function import Function

from nemo.collections.llm.gpt.model.megatron.hyena.hyena_config import HyenaConfig

try:
    from einops import rearrange
except ImportError:
    raise ImportError("einops is required by the Hyena model but cannot be imported")

try:
    from causal_conv1d import causal_conv1d_fn
except ImportError:

    def causal_conv1d_fn(*args, **kwargs):
        """Not imported: causal_conv1d_fn. An error will be raised if this is called."""
        raise ImportError("causal_conv1d is required by the Hyena model but cannot be imported")


try:
    from subquadratic_ops_torch.b2b_causal_conv1d import b2b_causal_conv1d
    from subquadratic_ops_torch.causal_conv1d import causal_conv1d
    from subquadratic_ops_torch.fft_causal_conv1d import fft_causal_conv1d
    from subquadratic_ops_torch.implicit_filter import implicit_filter
except ImportError:

    def causal_conv1d(*args, **kwargs):
        """Not imported: causal_conv1d. An error will be raised if this is called."""
        raise ImportError("subquadratic_ops not installed. causal_conv1d is not available.")

    def b2b_causal_conv1d(*args, **kwargs):
        """Not imported: b2b_causal_conv1d. An error will be raised if this is called."""
        raise ImportError("subquadratic_ops not installed. b2b_causal_conv1d is not available.")

    def fft_causal_conv1d(*args, **kwargs):
        """Not imported: fft_causal_conv1d. An error will be raised if this is called."""
        raise ImportError("subquadratic_ops not installed. fft_causal_conv1d is not available.")

    def implicit_filter(*args, **kwargs):
        """Not imported: implicit_filter. An error will be raised if this is called."""
        raise ImportError("subquadratic_ops not installed. implicit_filter is not available.")


def _get_zigzag_indices(N, device=None):
    """Generates the zigzag indices for rearrangement.

    Args:
        N (int): The total number of chunks.
        device (torch.device): The device on which to create tensors.

    Returns:
        torch.Tensor: The zigzag indices.
    """
    half_N = (N + 1) // 2
    idx1 = torch.arange(half_N, device=device)
    idx2 = torch.arange(N - 1, half_N - 1, -1, device=device)
    zigzag_idx = torch.empty(N, dtype=torch.long, device=device)
    zigzag_idx[0::2] = idx1
    zigzag_idx[1::2] = idx2
    return zigzag_idx


def _get_inverse_zigzag_indices(N, device=None):
    """Generates the inverse zigzag indices for rearrangement.

    Args:
        N (int): The total number of chunks.
        device (torch.device): The device on which to create tensors.

    Returns:
        torch.Tensor: The inverse zigzag indices.
    """
    half_N = N // 2
    idx1 = torch.arange(half_N, device=device)
    idx2 = torch.arange(N - 1, half_N - 1, -1, device=device)
    zigzag_idx = torch.empty(N, dtype=torch.long, device=device)
    zigzag_idx[0::2] = idx1
    zigzag_idx[1::2] = idx2
    inverse_zigzag_idx = torch.argsort(zigzag_idx)
    return inverse_zigzag_idx


def all_to_all_single_fn(
    group: dist.ProcessGroup,
    type: Literal["split_to_full", "full_to_split"],
    input: torch.Tensor,
    with_zigzag_splitting: bool = True,
) -> torch.Tensor:
    """Autograd-aware all_to_all_single communication function.

    Args:
        group (dist.ProcessGroup): The process group for communication.
        type (str): Either 'split_to_full' or 'full_to_split' to specify the communication pattern.
        input (torch.Tensor): Input tensor to be communicated.
        with_zigzag_splitting (bool, optional): Whether to apply zigzag splitting. Defaults to True.

    Returns:
        torch.Tensor: Output tensor after communication.
    """
    world_size = dist.get_world_size(group=group)

    if type == "split_to_full":
        """Given an split sequence, it gathers the whole sequence, while splitting across the channels dimension."""

        B, D, local_length = input.shape
        L = local_length * world_size
        d = D // world_size

        # Reshape and permute input for communication
        input_reshaped = rearrange(
            input, "B (cp d) l -> cp B d l", cp=world_size
        ).contiguous()  # [cp_world_size, B, d, l]

        # Perform all_to_all_single communication
        output_reshaped = torch.empty_like(input_reshaped)
        dist.all_to_all_single(output_reshaped, input_reshaped, group=group)  # [cp_world_size, B, d, l]

        # Permute and reshape output back to original form
        output = rearrange(output_reshaped, "cp B d l -> B d (cp l)", cp=world_size).contiguous()

        if with_zigzag_splitting:
            num_chunks = 2 * world_size
            unzigzagged_split_length = L // num_chunks  # Length of each small chunk
            device = output.device
            inverse_zigzag_idx = _get_inverse_zigzag_indices(num_chunks, device=device)

            # Vectorized rearrangement using inverse zigzag indices
            output = (
                output.reshape(B, d, num_chunks, unzigzagged_split_length)
                .index_select(dim=-2, index=inverse_zigzag_idx)
                .reshape(B, d, L)
            )

        return output

    elif type == "full_to_split":
        """
        Given a full sequence split across channels, splits across the sequence length while gathering the channels.
        """

        B, d, L = input.shape

        if with_zigzag_splitting:
            num_chunks = 2 * world_size
            chunk_length = L // num_chunks  # Length of each small chunk
            device = input.device
            zigzag_idx = _get_zigzag_indices(num_chunks, device=device)

            # Ensure L is divisible by num_chunks
            if L % num_chunks != 0:
                raise ValueError(f"Sequence length {L} is not divisible by num_chunks {num_chunks}")

            # Vectorized rearrangement using zigzag indices
            input = (
                input.reshape(B, d, num_chunks, chunk_length).index_select(dim=-2, index=zigzag_idx).reshape(B, d, L)
            )

        # Reshape and permute inputs for communication
        input_reshaped = rearrange(
            input, "b d (cp l) -> cp b d l", cp=world_size
        ).contiguous()  # [cp_world_size, b, d, l]

        # Perform all_to_all_single communication
        output_reshaped = torch.empty_like(input_reshaped)
        dist.all_to_all_single(output_reshaped, input_reshaped, group=group)  # [cp_world_size, B, d, l]

        # Permute and reshape outputs back to original form
        output = rearrange(output_reshaped, "cp b d l -> b (cp d) l", cp=world_size).contiguous()

        return output

    else:
        raise ValueError(f"Unknown type {type}")


class AllToAllSingleFunction(Function):
    """A custom autograd function for performing all_to_all_single communication with optional zigzag splitting.

    Attributes:
    - ctx: A context object that stores information for the forward and backward passes.
    - group: The process group for communication.
    - type: The type of communication pattern ('split_to_full' or 'full_to_split').
    - with_zigzag_splitting: A boolean indicating whether to apply zigzag splitting.
    """

    @staticmethod
    def forward(ctx, input_tensor, group, type, with_zigzag_splitting):
        """Forward pass for the AllToAllSingleFunction."""
        ctx.group = group
        ctx.type = type
        ctx.with_zigzag_splitting = with_zigzag_splitting

        # Detach input_tensor to prevent PyTorch from tracking operations inside the communication
        input_tensor = input_tensor.detach()

        # Perform the communication operation
        output = all_to_all_single_fn(
            group=ctx.group, type=ctx.type, input=input_tensor, with_zigzag_splitting=ctx.with_zigzag_splitting
        )

        return output

    @staticmethod
    def backward(ctx, grad_output):
        """Backward pass for the AllToAllSingleFunction."""
        # The backward pass will perform the reverse communication
        grad_input = all_to_all_single_fn(
            group=ctx.group,
            type="split_to_full" if ctx.type != "split_to_full" else "full_to_split",
            input=grad_output,
            with_zigzag_splitting=ctx.with_zigzag_splitting,
        )
        # Return the gradient w.r.t. the input_tensor and None for other arguments
        return grad_input, None, None, None


def zigzag_get_overlapping_patches(data, seq_dim, overlap_size):
    """Extracts the overlapping patches from data in each rank.

    Arguments:
        data (torch.Tensor): The concatenated data (chunk_a and chunk_b), e.g., [0, 3] , [1, 2] with zigzag and 2 GPUs.
        seq_dim (int): The sequence dimension along which the data is concatenated.
        overlap_size (int): The size of the overlapping patch.

    Returns:
        overlap_a, overlap_b (torch.Tensor): The overlapping chunks from the data. That is the end of the lowest, and
        the beginning of the last, e.g., end for 0 and start for 3.
    """
    assert seq_dim >= 0, "Negative indexes not supported."

    data_shape = list(data.shape)
    modified_shape = list(data.shape)
    modified_shape[seq_dim : seq_dim + 1] = [2, data_shape[seq_dim] // 2]

    reshaped_data = torch.reshape(data, modified_shape)

    # Move the dimension of the chunks to the first position
    # Create a permutation where seq_dim is moved to position 0
    permute_order = list(range(len(reshaped_data.shape)))
    permute_order.insert(0, permute_order.pop(seq_dim))  # Move seq_dim to index 0

    reshaped_data = reshaped_data.permute(dims=permute_order)

    seq_len = reshaped_data.shape[seq_dim + 1]  # Remember that a new dimension was added.
    overlapping_patches = reshaped_data.narrow(
        dim=seq_dim + 1, start=seq_len - overlap_size, length=overlap_size
    )  # Last n elements.
    return overlapping_patches[0], overlapping_patches[1]


class ExchangeOverlappingRegionsCausal(Function):
    """A custom autograd function for exchanging overlapping regions between chunks of data in a causal manner.

    The data is split across multiple GPUs using a distributed process group. The forward method handles the exchange
    of overlapping regions between chunks, while the backward method computes the gradients.

    Attributes:
    - ctx: A context object that stores information for the forward and backward passes.
    - chunk_a: Chunk to pass to the left.
    - chunk_b: Chunk to pass to the right.
    - group: The CP group
    - group_rank: The rank in the cp_group.
    """

    @staticmethod
    def forward(ctx, chunk_a, chunk_b, group, group_rank):
        """Forward pass for the ExchangeOverlappingRegionsCausal function."""
        group_ranks = dist.get_process_group_ranks(group)  # Get all global ranks in the cp_group
        group_world_size = len(group_ranks)  # Size of the cp_group

        ctx.group = group
        ctx.group_rank = group_rank
        ctx.group_world_size = group_world_size
        ctx.group_ranks = group_ranks

        # Initialize requests
        reqs = []

        # Exchange overlaps for chunk_a
        if group_rank > 0:
            # Receive overlap from previous rank
            recv_shape = list(chunk_a.shape)
            recv_prev_a = torch.empty(recv_shape, dtype=chunk_a.dtype, device=chunk_a.device)
            req_recv_a = dist.irecv(recv_prev_a, src=group_ranks[group_rank - 1])
            reqs.append(req_recv_a)
        else:
            recv_prev_a = None

        if group_rank < group_world_size - 1:
            # Send overlap to next rank
            req_send_a = dist.isend(chunk_a.contiguous(), dst=group_ranks[group_rank + 1])
            reqs.append(req_send_a)

        # Exchange overlaps for chunk_b
        if group_rank < group_world_size - 1:
            # Receive overlap from next rank
            recv_shape = list(chunk_b.shape)
            recv_next_b = torch.empty(recv_shape, dtype=chunk_b.dtype, device=chunk_b.device)
            req_recv_b = dist.irecv(recv_next_b, src=group_ranks[group_rank + 1])
            reqs.append(req_recv_b)
        else:
            recv_next_b = None

        if group_rank > 0:
            # Send overlap to previous rank
            req_send_b = dist.isend(chunk_b.contiguous(), dst=group_ranks[group_rank - 1])
            reqs.append(req_send_b)

        # Wait for all communication to finish
        for req in reqs:
            req.wait()

        # If no chunks received, use zeros instead (for consistency)
        if recv_prev_a is None:
            recv_prev_a = torch.zeros_like(chunk_a, dtype=chunk_a.dtype, device=chunk_a.device)
        if recv_next_b is None:
            recv_next_b = chunk_a.clone().contiguous()  # Got to receive from the same rank, but previous split.

        return recv_prev_a, recv_next_b

    @staticmethod
    def backward(ctx, grad_chunk_a, grad_chunk_b):
        """Backward pass for the ExchangeOverlappingRegionsCausal function."""
        # chunk_a, chunk_b = ctx.saved_tensors
        group_rank = ctx.group_rank
        group_world_size = ctx.group_world_size
        group_ranks = ctx.group_ranks

        # Initialize gradients with zeros
        _grad_chunk_a = torch.zeros_like(grad_chunk_a)
        _grad_chunk_b = torch.zeros_like(grad_chunk_b)

        # Initialize requests
        reqs = []

        # Handling grad_chunk_a

        # If rank > 0, send grad_recv_prev_a to rank - 1
        if group_rank > 0:
            req_send_a = dist.isend(grad_chunk_a.contiguous(), dst=group_ranks[group_rank - 1])
            reqs.append(req_send_a)
        else:
            # At rank 0, there's no previous rank to receive from, so we only consider local gradient contributions
            pass  # No action needed

        # If rank < world_size - 1, receive grad_chunk_a from rank + 1
        if group_rank < group_world_size - 1:
            grad_chunk_a_recv = torch.empty_like(grad_chunk_a)
            req_recv_a = dist.irecv(grad_chunk_a_recv, src=group_ranks[group_rank + 1])
            reqs.append(req_recv_a)

        # Handling grad_chunk_b

        # If rank < world_size - 1, send grad_recv_next_b to rank + 1
        if group_rank < group_world_size - 1:
            req_send_b = dist.isend(grad_chunk_b.contiguous(), dst=group_ranks[group_rank + 1])
            reqs.append(req_send_b)

        # If rank > 0, receive grad_chunk_b from rank - 1
        if group_rank > 0:
            grad_chunk_b_recv = torch.empty_like(grad_chunk_b)
            req_recv_b = dist.irecv(grad_chunk_b_recv, src=group_ranks[group_rank - 1])
            reqs.append(req_recv_b)

        # Wait for all communication to finish
        for req in reqs:
            req.wait()

        # Add received gradients
        if group_rank < group_world_size - 1:
            _grad_chunk_a = grad_chunk_a_recv

        if group_rank > 0:
            _grad_chunk_b = grad_chunk_b_recv

        if group_rank == group_world_size - 1:
            _grad_chunk_a = grad_chunk_b  # In the last split, the chunks are exchanged locally.

        return _grad_chunk_a, _grad_chunk_b, None, None, None


# End of CP related functions


def hyena_no_weight_decay_cond(name, param):
    """Condition for no weight decay for Hyena parameters."""
    # ImplicitModalFilter parameters
    if name.endswith('filter.p') or name.endswith('filter.R') or name.endswith('filter.gamma'):
        no_wd = True

    # ExplicitSingleDecayFilter parameters
    elif name.endswith('filter.h'):
        no_wd = True

    # TODO: Add overrides for other filter types if needed
    #       Alternatively - do something broader, like `if '.filter.' in name` ???

    # ParallelShortHyenaOperator parameters --> The parameters of the internal ParallelCausalDepthwiseConv1d object
    elif name.endswith('short_conv.short_conv_weight'):
        no_wd = True

    # All other parameters - use default MCore behavior:
    # Do not regularize biases and norm parameters
    # (See megatron.core.optimizer._get_pram_groups)
    else:
        no_wd = name.endswith(".bias") or len(param.shape) == 1

    return no_wd


def fftconv_func(u, k, D, dropout_mask, gelu=True, k_rev=None, bidirectional=False, use_subquadratic_ops=False):
    """Apply a 1D convolution to the input sequence u using the filter k and the shortcut D."""
    seqlen = u.shape[-1]
    fft_size = 2 * seqlen

    # check if k is less than seqlen -- subquadratic_ops input does not need padding
    if not use_subquadratic_ops and k.shape[-1] < seqlen:
        # Pad the filter k to the length of the input sequence u
        k = torch.nn.functional.pad(k, (0, seqlen - k.shape[-1]))

    # bidirectional
    if bidirectional:
        u_f = torch.fft.rfft(u.to(dtype=k.dtype), n=fft_size)

        # split k along the channel dimension
        k, k2 = k.split(k.shape[1] // 2, dim=1)

        # get fft of both k's
        k_f = torch.fft.rfft(k, n=fft_size) / fft_size
        k2_f = torch.fft.rfft(k2, n=fft_size) / fft_size

        if len(u.shape) > 3:
            k_f = k_f.unsqueeze(1)
            k2_f = k2_f.unsqueeze(1)

        y1 = u_f * k_f
        y2 = u_f.conj() * k2_f.conj()

        y = torch.fft.irfft(y1 + y2, n=fft_size, norm="forward")[..., :seqlen]

    # causal
    else:
        if use_subquadratic_ops:
            y = fft_causal_conv1d(u, k.squeeze(0))
        else:
            k_f = torch.fft.rfft(k, n=fft_size) / fft_size
            if k_rev is not None:
                k_rev_f = torch.fft.rfft(k_rev, n=fft_size) / fft_size
                k_f = k_f + k_rev_f.conj()

            u_f = torch.fft.rfft(u.to(dtype=k.dtype), n=fft_size)

            if len(u.shape) > 3:
                k_f = k_f.unsqueeze(1)

            y = torch.fft.irfft(u_f * k_f, n=fft_size, norm="forward")[..., :seqlen]

    out = y + u * D.unsqueeze(-1)
    if gelu:
        out = F.gelu(out)
    if dropout_mask is not None:
        return (out * rearrange(dropout_mask, "b H -> b H 1")).to(dtype=u.dtype)
    else:
        return out.to(dtype=u.dtype)


class ImplicitModalFilter(nn.Module):
    """An implicit modal filter."""

    def __init__(
        self,
        d_model,
        order=64,
        L_cache=None,
        gamma_min=0.01,
        gamma_max=0.1,
        lr=None,
        device=None,
        use_subquadratic_ops=False,
    ):
        super().__init__()
        self.order = order
        self.d_model = d_model
        self.use_subquadratic_ops = use_subquadratic_ops

        # Determine device safely
        if device is None:
            if torch.cuda.is_available():
                device = torch.cuda.current_device()
            else:
                device = torch.device('cpu')

        self.device = device

        if self.use_subquadratic_ops:
            self.implicit_filter = implicit_filter
            self.t = None
        else:
            # Do not register into buffer, so it doesn't cast to BF16!
            self.t = torch.arange(L_cache, dtype=torch.float32, device=self.device).view(1, 1, -1)  # 1, 1, L_cache

        with get_cuda_rng_tracker().fork():
            gamma = (
                torch.rand(self.d_model, order, dtype=torch.float32, device=self.device) * (gamma_max - gamma_min)
                + gamma_min
            )
            gamma = gamma.log()
            self.gamma = nn.Parameter(gamma)

            R = 1e-1 * torch.randn(d_model, order, dtype=torch.float32, device=self.device) / math.sqrt(order)
            self.R = nn.Parameter(R)
            self.p = nn.Parameter(-torch.ones(d_model, order, dtype=torch.float32, device=self.device))
            setattr(self.gamma, 'tensor_model_parallel', True)
            setattr(self.R, 'tensor_model_parallel', True)
            setattr(self.p, 'tensor_model_parallel', True)

    def get_t(self, L: int) -> torch.Tensor:
        """Get the t tensor."""
        if self.t is not None and self.t.shape[-1] >= L:
            # When L <= maximum previously requested length, we can take a subset of the cached t.
            return self.t[..., :L]
        else:
            # We are requesting an L that is longer than the cached t, grow t to the requested length.
            t = torch.arange(L, dtype=torch.float32, device=self.device).view(1, 1, -1)  # 1, 1, L
            self.t = t
            return self.t

    def get_logp(self):
        """Compute the log poles for the implicit modal filter."""
        logp = -torch.exp(self.p.to(torch.float32))
        glogp = logp * torch.exp(self.gamma.to(torch.float32))
        return glogp

    def compute_filter(self, L, t):
        """Compute the filter for convolution."""
        assert (
            t.dtype == torch.float32
        ), f"t must be float32. At lower precision, indexes will be merged together. Current dtype: {t.dtype}"
        # TODO: See if we can get this kernel to stay FP32. We can but it does not work with the distributed optimizer.
        # assert (
        #     self.p.dtype == torch.float32
        # ), f"p must be float32. At lower precision, indexes will be merged together. Current dtype: {self.p.dtype}"
        # assert (
        #     self.gamma.dtype == torch.float32
        # ), ("gamma must be float32. At lower precision, indexes will be merged together. "
        #    f"Current dtype: {self.gamma.dtype}")
        # assert (
        #     self.R.dtype == torch.float32
        # ), f"R must be float32. At lower precision, indexes will be merged together. Current dtype: {self.R.dtype}"

        logp = -torch.exp(self.p.to(torch.float32))
        glogp = logp * torch.exp(self.gamma.to(torch.float32))
        h = torch.exp(glogp[..., None] * t)
        h = torch.einsum('do,dot->dt', self.R.to(torch.float32), h)
        h = h[None]

        return h, None

    def filter(self, L, *args, **kwargs):
        """Get t and the convolution filter for t and the requested sequence length."""
        if self.use_subquadratic_ops:
            h = self.implicit_filter(self.get_logp(), self.R.to(torch.float32), L)
            h = h.unsqueeze(0)  # TODO: Remove this once we have a proper kernel implementation
        else:
            t = self.get_t(L)
            h = self.compute_filter(L, t)
        return h

    def forward(self, L, **kwargs):
        """Return the final convolutional filter for the requested sequence length."""
        return self.filter(L)

    def sharded_state_dict(self, prefix='', sharded_offsets=(), metadata=None):
        """Sharding along axis 0, bias not sharded."""
        state_dict = self.state_dict(prefix='', keep_vars=True)
        return make_sharded_tensors_for_checkpoint(state_dict, prefix, {'gamma': 0, 'R': 0, 'p': 0}, sharded_offsets)


class ExplicitSingleDecayFilter(nn.Module):
    """An explicit single decay filter."""

    def __init__(
        self,
        d_model,
        L_cache,
        log_r_min=0,
        log_r_max=2,
        unit_passthrough=False,
        decay_preset="strong",
        small_init=True,
        num_decay_repeats=1,
        device=None,
    ):
        super().__init__()

        # Determine device safely
        if device is None:
            if torch.cuda.is_available():
                device = torch.cuda.current_device()
            else:
                device = torch.device('cpu')

        self.device = device
        with get_cuda_rng_tracker().fork():
            h = torch.randn(d_model, L_cache, device=self.device) / math.sqrt(L_cache)
        assert decay_preset in ["strong", "normal", "weak"]
        if decay_preset == "strong":
            log_r_min = 0
            log_r_max = 2
        elif decay_preset == "normal":
            log_r_min = -1
            log_r_max = 2
        elif decay_preset == "weak":
            log_r_min = -2
            log_r_max = 2

        if small_init:
            h = h * 1e-5
        if unit_passthrough:
            h[:, :1] = 1.0
        self.num_decay_repeats = num_decay_repeats
        self.h = nn.Parameter(h)
        t = torch.linspace(0, 1, L_cache, device=self.device)[None]
        self.log_r_min = log_r_min
        self.log_r_max = log_r_max
        self.model_parallel_rank = get_tensor_model_parallel_rank()
        self.model_parallel_size = get_tensor_model_parallel_world_size()
        global_d_model = d_model * self.model_parallel_size // self.num_decay_repeats
        decay_domain = torch.logspace(log_r_min, log_r_max, global_d_model, device=self.device)[:, None].repeat(
            self.num_decay_repeats, 1
        )
        decay_domain = decay_domain[self.model_parallel_rank * d_model : (self.model_parallel_rank + 1) * d_model, :]
        decay = torch.exp(-decay_domain * t)
        self.register_buffer("decay", decay)
        setattr(self.h, 'tensor_model_parallel', True)
        setattr(self.decay, 'tensor_model_parallel', True)

    def forward(self, L, *args, **kwargs):
        """Forward pass for the explicit single decay filter.

        This returns the filter for the requested sequence length.
        """
        return self.filter(L, *args, **kwargs)

    @torch.compile(mode="max-autotune")
    def filter(self, L, *args, **kwargs):
        """Compute the filter as a function of h and decay for the requested sequence length."""
        h = self.h[:, :L]
        h = h * self.decay[:, :L]
        return h

    def sharded_state_dict(self, prefix='', sharded_offsets=(), metadata=None):
        """Sharding along axis 0, bias not sharded."""
        state_dict = self.state_dict(prefix='', keep_vars=True)
        return make_sharded_tensors_for_checkpoint(
            state_dict,
            prefix,
            {
                'h': 0,
                'decay': 0,
            },
            sharded_offsets,
        )


def small_init_init_method(dim):
    """Fills the input Tensor with values according to the method described in Transformers without Tears.

    Improving the Normalization of Self-Attention - Nguyen, T. & Salazar, J. (2010), using a normal distribution.
    """
    std = math.sqrt(2 / (5 * dim))

    def init_(tensor):
        return torch.nn.init.normal_(tensor, mean=0.0, std=std)

    return init_


def wang_init_method(n_layers, dim):
    """Initialize the weights of the model using the Wang initialization method."""
    std = 2 / n_layers / math.sqrt(dim)

    def init_(tensor):
        return torch.nn.init.normal_(tensor, mean=0.0, std=std)

    return init_


def get_init_method(init_method_name, num_layers, hidden_size):
    """Gets parameter initialization methods for the linear layers of the model."""
    if init_method_name == "wang_init":
        return wang_init_method(num_layers, hidden_size)
    elif init_method_name == "small_init":
        return small_init_init_method(hidden_size)
    else:
        raise NotImplementedError(f"Unknown init method {init_method_name}")


def ensure_divisibility(numerator, denominator):
    """Ensure that numerator is divisible by the denominator."""
    assert numerator % denominator == 0, "{} is not divisible by {}".format(numerator, denominator)


def divide(numerator, denominator):
    """Ensure that numerator is divisible by the denominator and return the division value."""
    ensure_divisibility(numerator, denominator)
    return numerator // denominator


def initialize_affine_weight_gpu(weight, init_method, partition_dim, stride=1):
    """Initialize affine weight for model parallel on GPU."""
    weight.model_parallel = True
    weight.partition_dim = partition_dim
    weight.partition_stride = stride

    with get_cuda_rng_tracker().fork():
        init_method(weight.data)  # modify the data in place


def get_groups_and_group_sizes(hidden_size, num_groups, world_size, expand_factor):
    """Get the groups and group sizes for the model."""
    width_per_tp_group = divide(hidden_size, world_size)
    num_groups_per_tp = int(divide(num_groups, world_size) * expand_factor)
    group_dim = width_per_tp_group // num_groups_per_tp
    return width_per_tp_group, num_groups_per_tp, group_dim


class ParallelHyenaOperator(nn.Module):
    """A class for the ParallelHyenaOperator."""

    def __init__(
        self,
        hidden_size,
        transformer_config: TransformerConfig,
        hyena_config: HyenaConfig,
        init_method,
        operator_type,
        max_sequence_length,
        zigzag=True,
    ):
        super().__init__()

        # Determine device safely
        if torch.cuda.is_available():
            self.device = torch.cuda.current_device()
        else:
            self.device = torch.device('cpu')

        self.hidden_size = hidden_size
        self.transformer_config = transformer_config
        self.hyena_config = hyena_config
        self.operator_type = operator_type
        self.fp16 = transformer_config.fp16
        self.bf16 = transformer_config.bf16
        self.use_conv_bias = True  # Always True. Added here for compatibility with B2BCausalConv1dModule.

        if self.operator_type == "hyena_medium_conv" and hyena_config.hyena_medium_filter_cls is not None:
            self.hyena_filter_cls = hyena_config.hyena_medium_filter_cls
        else:
            self.hyena_filter_cls = hyena_config.hyena_filter_cls

        self.bidirectional = hyena_config.bidirectional
        self.use_hyena_filter = hyena_config.use_hyena_filter
        self.zigzag = zigzag

        self.use_subquadratic_ops = transformer_config.use_subquadratic_ops

        self.model_parallel_size = get_tensor_model_parallel_world_size()
        self.model_parallel_rank = get_tensor_model_parallel_rank()

        self.L = max_sequence_length

        if self.operator_type == "hyena_medium_conv":
            self.num_groups = (
                hyena_config.num_groups_hyena_medium
                if hyena_config.num_groups_hyena_medium is not None
                else hyena_config.num_groups_hyena
            )
        elif self.operator_type == "hyena_short_conv":
            self.num_groups = (
                hyena_config.num_groups_hyena_short
                if hyena_config.num_groups_hyena_short is not None
                else hyena_config.num_groups_hyena
            )
        else:
            # default to the global num_groups_hyena
            self.num_groups = hyena_config.num_groups_hyena

        if self.num_groups is None:
            self.num_groups = transformer_config.hidden_size

        world_size: int = get_tensor_model_parallel_world_size()

        self.width_per_tp_group, self.num_groups, self.group_dim = get_groups_and_group_sizes(
            self.hidden_size, self.num_groups, world_size, hyena_config.hyena_width_expansion
        )

        self.short_conv_L = hyena_config.short_conv_L
        self.use_medium_hyena = True if self.operator_type == "hyena_medium_conv" else False
        self.hyena_medium_conv_len = hyena_config.hyena_medium_conv_len

        # TODO: Check which of these filters can be removed
        #       At the moment only "explicit_single_decay" and "implicit_modal" are used
        if self.hyena_filter_cls == "explicit_single_decay":
            self.filter = ExplicitSingleDecayFilter(
                d_model=self.num_groups,
                L_cache=self.hyena_medium_conv_len,
                decay_preset=hyena_config.explicit_filter_decay_preset,
                device=self.device,
            )
            self.kernel_size = self.hyena_medium_conv_len
        elif self.hyena_filter_cls == "implicit_modal":
            self.filter = ImplicitModalFilter(
                d_model=self.num_groups,
                L_cache=self.L,
                order=hyena_config.hyena_filter_order,
                gamma_min=hyena_config.modal_gamma_min,
                gamma_max=hyena_config.modal_gamma_max,
                use_subquadratic_ops=self.use_subquadratic_ops,
                device=self.device,
            )
            self.kernel_size = self.L
        else:
            raise ValueError(f"Unknown hyena filter class: {self.hyena_filter_cls}")

        with get_cuda_rng_tracker().fork():
            self.conv_bias = nn.Parameter(
                torch.empty(
                    self.width_per_tp_group,
                    device=self.device,
                    dtype=transformer_config.params_dtype,
                )
            )
            # Add attribute to prevent automatic casting during model conversion
            setattr(self.conv_bias, 'tensor_model_parallel', True)
            bounds = math.sqrt(1 / self.kernel_size)
            conv_init_method = partial(torch.nn.init.uniform_, a=-bounds, b=bounds)
            self.conv_bias.data = conv_init_method(self.conv_bias.data)
            self.conv_bias.model_parallel = True
            self.conv_bias.partition_dim = 0
            self.conv_bias.stride = 1

    def forward_long(self, *, x1, x2, v, h, bias, inference_context):
        """Forward pass long."""
        import nemo.collections.llm.gpt.model.megatron.hyena.engine as engine

        def update_filter_state(filter_name, *, state):
            if not inference_context:
                return
            key = f"{filter_name}_filter_state_dict"
            filter_state_dict = getattr(inference_context, key, {})
            filter_state_dict[id(self)] = state
            setattr(inference_context, key, filter_state_dict)

        def get_filter_state(filter_name):
            key = f"{filter_name}_filter_state_dict"
            return getattr(inference_context, key, {}).get(id(self))

        # x1, x2, v all of shape torch.Size([1, 4096, 63])
        u = torch.cat([x2, x1, v], dim=1)  # torch.Size([1, 12288, 63])
        L = u.shape[-1]
        poles = rearrange(self.filter.p, "d n -> d n 1")  # n = 16
        poles = self.filter.get_logp()
        poles = rearrange(poles, "d n -> d n 1")  # n = 16

        iir_state = get_filter_state("iir")
        if iir_state is None:
            y, iir_state = engine.parallel_iir(
                z_pre=u,  # [1 d l]
                h=h,  # must be in [1 d l]
                D=bias,  # self.short_filter_bias,
                L=L,
                poles=poles,
                t=self.filter.get_t(L),  # torch.Size([1, 1, L])
                hidden_size=self.hidden_size,
                compute_state=inference_context is not None,
            )
            # y = rearrange(y, "b d l -> b l d")
        else:
            x1 = rearrange(x1, "1 d l -> l d")
            x2 = rearrange(x2, "1 d l -> l d")
            v = rearrange(v, "1 d l -> l d")
            x1, x2 = x2, x1  # TODO: figure why it is swapped
            y, iir_state = engine.step_iir(
                x2=x2,
                x1=x1,
                v=v,
                D=bias,  # torch.Size([4096])
                residues=self.filter.R,  # torch.Size([4096, 16])
                poles=poles,  # torch.Size([4096, 16, 1])
                iir_state=iir_state,
            )
            y = rearrange(y, "b d -> b 1 d")
            y = y.to(dtype=x1.dtype)
        update_filter_state("iir", state=iir_state)
        return rearrange(y, "b l d -> b d l")  # b l d

    def forward_medium(self, *, x1, x2, v, h, bias, inference_context):
        """Forward pass medium."""
        import nemo.collections.llm.gpt.model.megatron.hyena.engine as engine

        def update_filter_state(filter_name, *, state):
            if not inference_context:
                return
            key = f"{filter_name}_filter_state_dict"
            filter_state_dict = getattr(inference_context, key, {})
            filter_state_dict[id(self)] = state
            setattr(inference_context, key, filter_state_dict)

        def get_filter_state(filter_name):
            key = f"{filter_name}_filter_state_dict"
            return getattr(inference_context, key, {}).get(id(self))

        x1, x2, v = rearrange([x1, x2, v], "h b d l -> h b l d")
        # all above in [B D L]
        u = x2 * v  # b l d
        L = u.shape[1]
        h = rearrange(h, "d l -> d 1 l")

        fir_state = get_filter_state("inner_fir")
        if fir_state is None:
            y, fir_state = engine.parallel_fir(
                u=u,
                weight=h,  # self.short_filter_weight,
                bias=bias,  # self.short_filter_bias,
                gated_bias=self.kernel_size >= 128,  # aka: only for medium filter
                L=L,
                fir_length=self.kernel_size,  # self.short_filter_length,
                compute_state=inference_context is not None,
            )
            y = rearrange(y, "b d l -> b l d")
            y = y * x1
        else:
            u = rearrange(u, "b 1 d -> b d")
            x1 = rearrange(x1, "b 1 d  -> b d")
            y, fir_state = engine.step_fir(
                u=u,
                fir_state=fir_state,
                weight=h,
                bias=bias,
                gated_bias=self.kernel_size >= 128,  # aka: only for medium filter
                flip_filter=self.kernel_size >= 128,  # aka: only for medium filter
            )
            y = x1 * y
            y = rearrange(y, "b d -> b 1 d")
        update_filter_state("inner_fir", state=fir_state)
        return rearrange(y, "b l d -> b d l")  # b l d

    def forward(self, x1, x2, v, _hyena_use_cp=True, inference_context=None):
        """Shape specification for inputs and outputs.

        Input shapes: bs, (num_groups, group_size), seq_length
        Output shapes: bs, (num_groups, group_size), seq_length
        """
        B, GDG, L = x1.shape
        x1, x2, v = x1[..., :L], x2[..., :L], v[..., :L]

        # CP control
        if _hyena_use_cp:
            cp_group = get_context_parallel_group()
        else:
            cp_group = None

        # The kernel length must be adjusted in CP settings
        _L_kernel = L if cp_group is None else L * len(torch.distributed.get_process_group_ranks(cp_group))
        if self.use_medium_hyena:
            h = self.filter(self.hyena_medium_conv_len)
        else:
            h = self.filter(_L_kernel)

        if isinstance(h, tuple):
            h = h[0]

        conv_bias = self.conv_bias
        local_size = None

        # Initialize z split for non-inference cases
        z = x2 * v

        if cp_group is not None and len(torch.distributed.get_process_group_ranks(cp_group)) > 1:

            if inference_context is not None:  # reconstruct ALL tensors from split to full
                x1, x2, v = [
                    AllToAllSingleFunction.apply(tensor, cp_group, "split_to_full", True) for tensor in [x1, x2, v]
                ]
            else:  # only reconstruct z (post gating) for non-inference case
                z = AllToAllSingleFunction.apply(z, cp_group, "split_to_full", True)

            # the tensors are now split across channels, but have full length.
            # [ B, H // num_ranks, L]

            rank = torch.distributed.get_rank(cp_group)
            local_size = self.num_groups // get_context_parallel_world_size()

            if isinstance(self.filter, (ImplicitModalFilter)):
                h = h[:, rank * local_size : (rank + 1) * local_size]
            elif isinstance(self.filter, ExplicitSingleDecayFilter):
                h = h[rank * local_size : (rank + 1) * local_size]
            else:
                raise ValueError(f"Kernels of type {self.filter.__class__} have not been verified with CP.")

            local_bias_size = self.width_per_tp_group // get_context_parallel_world_size()
            conv_bias = self.conv_bias[rank * local_bias_size : (rank + 1) * local_bias_size]

        h = h.repeat_interleave(self.group_dim, dim=-2)

        if inference_context is not None:  # Needs full length x1 x2 v
            if self.operator_type == "hyena_medium_conv":
                return self.forward_medium(x1=x1, x2=x2, v=v, h=h, bias=conv_bias, inference_context=inference_context)
            elif self.operator_type == "hyena":
                return self.forward_long(x1=x1, x2=x2, v=v, h=h, bias=conv_bias, inference_context=inference_context)
        else:  # Needs full length z (post gating)
            # with torch.autocast("cuda"):
            z = fftconv_func(
                u=z.to(torch.float32),
                k=h.to(torch.float32),
                D=conv_bias.to(torch.float32),
                dropout_mask=None,
                gelu=False,
                bidirectional=self.bidirectional,
                use_subquadratic_ops=self.use_subquadratic_ops,
            )
            z = z.to(v.dtype)

            if cp_group is not None and len(torch.distributed.get_process_group_ranks(cp_group)) > 1:
                z = AllToAllSingleFunction.apply(z, cp_group, "full_to_split", True)
                # [ B, H, L // num_ranks]

            z = x1 * z
            return z  # [B, (G, DG), L]

    def sharded_state_dict(self, prefix='', sharded_offsets=(), metadata=None):
        """Sharded state dictionary for the ParallelHyenaOperator."""
        sharded_state_dict = {}
        # Parameters
        self._save_to_state_dict(sharded_state_dict, '', keep_vars=True)
        sharded_state_dict = make_sharded_tensors_for_checkpoint(
            sharded_state_dict,
            prefix,
            tensor_parallel_layers_axis_map={
                'conv_bias': 0,
            },  # parameters sharded across TP
            sharded_offsets=sharded_offsets,
        )
        # Submodules
        for name, module in self.named_children():
            module_sharded_sd = sharded_state_dict_default(module, f'{prefix}{name}.', sharded_offsets, metadata)

            sharded_state_dict.update(module_sharded_sd)
        return sharded_state_dict


class ParallelShortHyenaOperator(nn.Module):
    """A class for the ParallelShortHyenaOperator."""

    def __init__(
        self,
        hidden_size,
        transformer_config: TransformerConfig,
        hyena_config: HyenaConfig,
        init_method,
        short_conv_class,
        use_fast_causal_conv=False,
        local_init=False,
        use_conv_bias=True,
    ):
        super().__init__()
        self.transformer_config = transformer_config
        self.hyena_config = hyena_config
        self.hidden_size = hidden_size
        self.use_fast_causal_conv = use_fast_causal_conv

        world_size: int = get_tensor_model_parallel_world_size() if not local_init else 1
        # assert, if using fast_conv_mixer, then the hyena_short_conv_len must be 3
        if use_fast_causal_conv and not transformer_config.use_subquadratic_ops:
            assert hyena_config.hyena_short_conv_len <= 4, "fast_conv_mixer requires hyena_short_conv_len <= 4"

        kernel_size = hyena_config.hyena_short_conv_len
        self.pregate = hyena_config.hyena_short_conv_pregate
        self.postgate = hyena_config.hyena_short_conv_postgate
        self.num_groups = (
            hyena_config.num_groups_hyena_short
            if hyena_config.num_groups_hyena_short is not None
            else hyena_config.num_groups_hyena
        )
        if self.num_groups is None:
            self.num_groups = transformer_config.hidden_size

        self.num_groups = int(self.num_groups * hyena_config.hyena_width_expansion)

        self.width_per_tp_group, self.num_groups, self.group_dim = get_groups_and_group_sizes(
            self.hidden_size, self.num_groups, world_size, hyena_config.hyena_width_expansion
        )

        self.short_conv = short_conv_class(
            self.width_per_tp_group,
            transformer_config,
            hyena_config=hyena_config,
            kernel_size=kernel_size,
            init_method=init_method,
            bias=hyena_config.conv_proj_bias,
            use_fast_causal_conv=use_fast_causal_conv,
            num_groups=self.num_groups,
            repeat_h_dg=False,
            local_init=local_init,
        )

        self.use_conv_bias = use_conv_bias
        if self.use_conv_bias:
            with get_cuda_rng_tracker().fork():
                self.conv_bias = nn.Parameter(
                    torch.empty(
                        self.num_groups,
                        device=torch.cuda.current_device(),
                        dtype=transformer_config.params_dtype,
                    )
                )
                setattr(self.conv_bias, 'tensor_model_parallel', True)
                bounds = math.sqrt(1 / kernel_size)
                conv_init_method = partial(torch.nn.init.uniform_, a=-bounds, b=bounds)
                self.conv_bias.data = conv_init_method(self.conv_bias.data)
                self.conv_bias.model_parallel = True
                self.conv_bias.partition_dim = 0
                self.conv_bias.stride = 1

    def forward(self, x1, x2, v, inference_context=None, _hyena_use_cp=True):
        """Shape specification for inputs and outputs.

        Input shapes: bs, (num_groups, group_size), seq_length
        Output shapes: bs, (num_groups, group_size), seq_length
        """
        B, GDG, L = x1.shape
        x1, x2, v = x1[..., :L], x2[..., :L], v[..., :L]

        z = x2 * v if self.pregate else v
        if not self.use_conv_bias:
            z = self.short_conv(z, _use_cp=_hyena_use_cp, inference_context=inference_context)
        else:
            # maybe handle num_groups
            bias = self.conv_bias.repeat_interleave(self.group_dim, dim=0)
            z = self.short_conv(z, _use_cp=_hyena_use_cp) + rearrange(bias, "h -> 1 h 1") * z  # conv(z) + bias * z

        z = x1 * z if self.postgate else z

        return z  # [B, (G, DG), L]

    def sharded_state_dict(self, prefix='', sharded_offsets=(), metadata=None):
        """Sharded state dictionary for the ParallelShortHyenaOperator."""
        sharded_state_dict = {}
        # Parameters
        self._save_to_state_dict(sharded_state_dict, '', keep_vars=True)
        sharded_state_dict = make_sharded_tensors_for_checkpoint(
            sharded_state_dict,
            prefix,
            tensor_parallel_layers_axis_map={
                'conv_bias': 0,
            },  # parameters sharded across TP
            sharded_offsets=sharded_offsets,
        )
        # Submodules
        for name, module in self.named_children():
            module_sharded_sd = sharded_state_dict_default(module, f'{prefix}{name}.', sharded_offsets, metadata)

            sharded_state_dict.update(module_sharded_sd)
        return sharded_state_dict


class ParallelCausalDepthwiseConv1d(nn.Module):
    """A class for the ParallelCausalDepthwiseConv1d."""

    def __init__(
        self,
        d_model,
        transformer_config: TransformerConfig,
        hyena_config: HyenaConfig,
        kernel_size,
        init_method,
        bias=False,  # not currently supported
        use_fast_causal_conv=False,
        num_groups=None,  # enables some weight sharing
        repeat_h_dg=True,
        local_init=False,
    ):
        super().__init__()
        self.d_model = d_model
        self.kernel_size = kernel_size
        self.use_conv_bias = bias
        self.use_fast_causal_conv = use_fast_causal_conv
        self.num_groups = num_groups
        self.transformer_config = transformer_config
        self.use_subquadratic_ops = transformer_config.use_subquadratic_ops

        if self.num_groups is None:
            self.num_groups = self.d_model

        self.group_dim = self.d_model // self.num_groups

        if self.use_fast_causal_conv and not self.use_subquadratic_ops:
            assert causal_conv1d_fn is not None, "custom causal conv not installed"
            weight_shape = [self.num_groups, kernel_size]
        elif (
            self.use_fast_causal_conv and self.use_subquadratic_ops
        ):  # hyena_proj_conv of LI layer when subquadratic_ops is enabled
            weight_shape = [self.num_groups, kernel_size]
        # use torch
        else:
            if hyena_config.use_depthwise_short_conv_grouping:
                weight_shape = [self.num_groups, 1, kernel_size]
                self.conv_groups = self.d_model

            else:
                if repeat_h_dg:
                    weight_shape = [self.num_groups, self.group_dim, kernel_size]
                else:
                    weight_shape = [self.num_groups, 1, kernel_size]

                self.conv_groups = self.num_groups

        with get_cuda_rng_tracker().fork():
            self.short_conv_weight = nn.Parameter(
                torch.empty(
                    weight_shape,
                    device=torch.cuda.current_device(),
                    dtype=transformer_config.params_dtype,
                )
            )
            setattr(self.short_conv_weight, 'tensor_model_parallel', True)

            # Use the standard PyTorch Conv1d class init:
            #   https://pytorch.org/docs/master/generated/torch.nn.Conv1d.html
            bounds = math.sqrt(1 / hyena_config.short_conv_L)
            conv_init_method = partial(torch.nn.init.uniform_, a=-bounds, b=bounds)
            if local_init:
                self.short_conv_weight.data = conv_init_method(self.short_conv_weight.data)
            else:
                # Call this on the module because it also modifies module attributes in addition to the data.
                initialize_affine_weight_gpu(self.short_conv_weight, conv_init_method, partition_dim=0)

    def forward(self, x, inference_context=None, _use_cp=True):
        """Forward pass for the ParallelCausalDepthwiseConv1d."""
        assert x.ndim == 3, "Only 3D tensors supported."

        x_shape = x.shape
        weight = self.short_conv_weight
        pad_size = self.kernel_size - 1

        # maybe handle num_groups
        weight = weight.repeat_interleave(self.group_dim, dim=0)

        if _use_cp and get_context_parallel_world_size() > 1:

            cp_group = get_context_parallel_group()
            cp_rank = get_context_parallel_rank()

            # Transfer patches across ranks.
            seq_dim = 2  # Last dimension.
            chunk_a, chunk_b = zigzag_get_overlapping_patches(x, seq_dim=seq_dim, overlap_size=pad_size)
            received_a, received_b = ExchangeOverlappingRegionsCausal.apply(chunk_a, chunk_b, cp_group, cp_rank)

            # Pad and rearrange
            x = rearrange(x, "b h (nc s) -> (nc b) h s", nc=2)
            padding = torch.concat([received_a, received_b], dim=0)

            x = torch.concat([padding, x], dim=-1)

        else:
            x = F.pad(x, (pad_size, 0))

        # subquadratic_ops causal_conv1d is only applied to the projection conv of Hyena LI layer
        # Projection conv is fused with SE/MR layers (B2BCausalConv1dModule)
        if self.use_fast_causal_conv:  # hyena_proj_conv case
            if self.use_subquadratic_ops:  # hyena_proj_conv of LI layer when subquadratic_ops is enabled
                y = causal_conv1d(x, weight)[..., pad_size:]
            else:
                y = causal_conv1d_fn(x, weight, bias=None, activation=None)[..., pad_size:]
        else:  # hyena_short_conv case

            y = F.conv1d(
                x,
                weight,
                bias=None,
                stride=1,
                padding=0,
                groups=self.conv_groups,
            )

        if _use_cp and get_context_parallel_world_size() > 1:
            y = rearrange(y, "(nc b) h s -> b h (nc s)", nc=2)

        assert y.shape == x_shape, f"y.shape = {y.shape} | x.shape = {x_shape}"

        return y

    def sharded_state_dict(self, prefix='', sharded_offsets=(), metadata=None):
        """Sharding along axis 0, bias not sharded."""
        state_dict = self.state_dict(prefix='', keep_vars=True)
        return make_sharded_tensors_for_checkpoint(
            state_dict,
            prefix,
            {
                'short_conv_weight': 0,
            },
            sharded_offsets,
        )


class B2BCausalConv1dModule(nn.Module):
    """Module that performs back-to-back causal convolution operations using optimized CUDA kernels.

    Combines the projection and mixer convolutions into a single optimized operation.

    Note: This module stores references to other modules without registering them as child modules
    to avoid duplicate parameters in the state dict. The actual parameters are owned by the
    parent HyenaMixer module's hyena_proj_conv and mixer attributes.
    """

    def __init__(
        self,
        proj_conv_module,
        mixer_module,
        operator_type="hyena_short_conv",
        b2b_causal_conv1d=b2b_causal_conv1d,
        flip_mixer_weight: bool = False,
    ):
        """Initialize the B2BCausalConv1dModule.

        Args:
            proj_conv_module: The projection convolution module that performs the initial projection
            mixer_module: The mixer module that performs the second convolution operation
            operator_type: The type of hyena operator to use, either "hyena_short_conv" or "hyena_medium_conv"
            b2b_causal_conv1d: The CUDA kernel function for optimized back-to-back causal convolution
            flip_mixer_weight: Whether to flip the mixer weight, when weights are coming from FFTConv mixer
        """
        super().__init__()
        self.b2b_causal_conv1d_fn = b2b_causal_conv1d
        # Store references to the modules WITHOUT registering them as child modules
        # Using object.__setattr__ bypasses PyTorch's module registration system
        # This prevents their parameters from appearing in the state dict with the b2b_kernel prefix
        object.__setattr__(self, '_proj_conv_module', proj_conv_module)
        object.__setattr__(self, '_mixer_module', mixer_module)
        self._use_conv_bias = self._mixer_module.use_conv_bias
        self.operator_type = operator_type
        self.flip_mixer_weight = flip_mixer_weight
        # Combined padding from both convolutions - this is a key difference from the
        # sequential execution of two convs which applies padding separately
        self._proj_conv_kernel_size = self._proj_conv_module.kernel_size
        if operator_type == "hyena_short_conv":
            # For short conv, the mixer module is ParallelCausalDepthwiseConv1d
            self._mixer_kernel_size = self._mixer_module.short_conv.kernel_size
        elif operator_type == "hyena_medium_conv":
            # For medium conv, we need to get the kernel size from the filter
            self._mixer_kernel_size = self._mixer_module.kernel_size
        else:
            raise ValueError(f"Operator type {operator_type} not supported")

        self.effective_pad_size = (self._mixer_kernel_size - 1) + (self._proj_conv_kernel_size - 1)

    def forward(self, x, _use_cp=True):
        """Forward pass for the B2BCausalConv1dModule.

        Args:
            x: Input tensor [B, D, L]
            _use_cp: Whether to use context parallelism
        Returns:
            Tensor: Output of the back-to-back causal convolution operation
        """
        # Validate input dimensions
        if x.dim() != 3:
            raise ValueError("Input tensor must be 3D [batch_size, hidden_dim, seq_len]")

        # Extract weights at runtime to avoid parameter registration
        proj_weight = self._proj_conv_module.short_conv_weight
        mixer_weight = None  # Initialize mixer_weight

        if self.operator_type == "hyena_short_conv":
            # For short conv, the mixer module is ParallelCausalDepthwiseConv1d
            mixer_weight = self._mixer_module.short_conv.short_conv_weight
        elif self.operator_type == "hyena_medium_conv":
            # For medium conv, we need to compute the filter weights first
            mixer_weight = self._mixer_module.filter.filter(self._mixer_module.hyena_medium_conv_len)
            if isinstance(mixer_weight, tuple):
                mixer_weight = mixer_weight[0]
            # The filter weights need to be in the shape [groups, 1, kernel_size]
            mixer_weight = mixer_weight.unsqueeze(1)  # Add channel dimension
            mixer_weight = mixer_weight.to(x.dtype)  # Convert to the same dtype as the input
        else:
            raise ValueError(f"Operator type {self.operator_type} not supported in B2BCausalConv1dModule")

        # Extract bias if needed
        if self._use_conv_bias:
            bias = self._mixer_module.conv_bias
            # The mixer bias should have shape [x.shape[1] // 3] for the CUDA kernel
            if bias.shape[0] != x.shape[1] // 3:
                # Expand the bias to match the input width
                bias = bias.repeat_interleave(x.shape[1] // (3 * bias.shape[0]), dim=0)
        else:
            # Create a zero bias with the correct shape [x.shape[1] // 3]
            bias = torch.zeros(x.shape[1] // 3, dtype=x.dtype, device=x.device)

        # Reshape proj_weight if needed (from [groups, channels, kernel_size] to [groups*channels, kernel_size])
        if proj_weight.dim() == 3:
            proj_weight = proj_weight.reshape(-1, proj_weight.size(-1))
        # Reshape mixer_weight if needed (from [groups, channels, kernel_size] to [groups*channels, kernel_size])
        if mixer_weight.dim() == 3:
            # If the middle dimension is 1, we can just squeeze it
            if mixer_weight.size(1) == 1:
                mixer_weight = mixer_weight.squeeze(1)
            else:
                # Otherwise reshape to flatten the first two dimensions
                mixer_weight = mixer_weight.reshape(-1, mixer_weight.size(-1))
        # For evo2 compatibility, since it uses FFTConv mixer and here we use direct convolution.
        if self.flip_mixer_weight:
            mixer_weight = torch.flip(mixer_weight, dims=[-1])

        # maybe handle num_groups
        proj_weight = proj_weight.repeat_interleave(self._proj_conv_module.group_dim, dim=0)
        mixer_weight = mixer_weight.repeat_interleave(self._mixer_module.group_dim, dim=0)

        # Support context parallelism similar to how it's done in ParallelCausalDepthwiseConv1d
        if _use_cp and get_context_parallel_world_size() > 1:
            # Validate sequence length for CP mode
            cp_size = get_context_parallel_world_size()
            if x.size(-1) % cp_size != 0:
                raise ValueError("Sequence length must be divisible by context parallel size")

            cp_group = get_context_parallel_group()
            cp_rank = get_context_parallel_rank()

            # Transfer patches across ranks
            seq_dim = 2  # Last dimension (L)

            # Get overlapping patches - using the combined effective padding size
            chunk_a, chunk_b = zigzag_get_overlapping_patches(x, seq_dim=seq_dim, overlap_size=self.effective_pad_size)

            # We're exchanging larger patches once instead of smaller patches twice
            received_a, received_b = ExchangeOverlappingRegionsCausal.apply(chunk_a, chunk_b, cp_group, cp_rank)

            # Pad and rearrange
            x = rearrange(x, "b h (nc s) -> (nc b) h s", nc=2)
            padding = torch.concat([received_a, received_b], dim=0)

            x = torch.concat([padding, x], dim=-1)  # [ncB, D, L]
            result = self.b2b_causal_conv1d_fn(x, proj_weight, mixer_weight, bias)
            result = result[..., self.effective_pad_size :]  # Remove padding from output
            result = rearrange(result, "(nc b) h s -> b h (nc s)", nc=2)
        else:
            # Call the CUDA kernel
            # Padding is not required and is handled in the CUDA kernel when CP=1 (padding=0)
            result = self.b2b_causal_conv1d_fn(x, proj_weight, mixer_weight, bias)
        return result


class ParallelCausalDepthwiseConv1dWithState(ParallelCausalDepthwiseConv1d):
    """A class for the ParallelCausalDepthwiseConv1dWithState."""

    def forward(self, x, inference_context=None, _use_cp=True):
        # If not in inference mode, use the original implementation
        if inference_context is None:
            return super().forward(x, _use_cp=_use_cp)
        features_BLD = rearrange(x, "b d l -> b l d").contiguous()
        u = features_BLD
        weight = self.short_conv_weight

        if len(weight.shape) == 2:
            weight = rearrange(weight, "hidden_size3 filter_len -> hidden_size3 1 filter_len")

        weight = weight.repeat_interleave(self.group_dim, dim=0)

        import nemo.collections.llm.gpt.model.megatron.hyena.engine as engine

        def update_filter_state(filter_name, *, state):
            if not inference_context:
                return
            key = f"{filter_name}_filter_state_dict"
            filter_state_dict = getattr(inference_context, key, {})
            filter_state_dict[id(self)] = state
            setattr(inference_context, key, filter_state_dict)

        def get_filter_state(filter_name):
            key = f"{filter_name}_filter_state_dict"
            return getattr(inference_context, key, {}).get(id(self))

        L = u.shape[1]
        fir_state = get_filter_state("fir")
        if fir_state is None:
            z_pre, fir_state = engine.parallel_fir(
                u=u,
                weight=torch.tensor(weight),  # self.short_filter_weight,
                bias=None,
                L=L,
                gated_bias=False,
                fir_length=self.kernel_size,  # self.short_filter_length,
                compute_state=inference_context is not None,
            )
        else:
            if len(u.shape) > 2:
                u = u[:, -1]

            z_pre, fir_state = engine.step_fir(
                u=u,
                fir_state=fir_state,
                weight=weight,
                bias=None,
            )
            z_pre = rearrange(z_pre, "b d -> b d 1")
        update_filter_state("fir", state=fir_state)
        return z_pre


def make_upper_case(tokens, lowercase_start=97, lowercase_end=122, case_diff=32):
    """Replace lowercase ASCII characters with uppercase.

    Args:
        tokens: Input tensor containing token IDs
        lowercase_start: ASCII value for the first lowercase character (default: 97 for 'a')
        lowercase_end: ASCII value for the last lowercase character (default: 122 for 'z')
        case_diff: Difference between lowercase and uppercase (default: 32)

    Returns:
        tuple: (uppercase_tensor, lowercase_mask)
    """
    lowercase_mask = (tokens >= lowercase_start) & (tokens <= lowercase_end)
    uppercase_tensor = torch.where(lowercase_mask, tokens - case_diff, tokens)
    return uppercase_tensor, lowercase_mask


def reweighted_cross_entropy(loss, labels, lowercase_weight=1.0, normalize_per_batch=True):
    """Modified for lower case loss reweighting, using the cross_entropy function as a base.

    If normalize_per_batch, loss_weights are normalized by the number of tokens in the batch so
    the magnitude of the loss is not affected by the number of upper/lower case letters
    otherwise, loss_weights are normalized by the number of tokens: combined_loss/len.

    Performs mean reduction and applies loss_mask.
    """
    labels, loss_mask, lowercase_mask = labels[0], labels[1], labels[2]

    upper_loss_mask = loss_mask.bool() & (~lowercase_mask.bool())
    lower_loss_mask = loss_mask.bool() & lowercase_mask.bool()

    loss_weights = torch.zeros_like(loss_mask)
    loss_weights[upper_loss_mask] = 1.0
    loss_weights[lower_loss_mask] = lowercase_weight

    if normalize_per_batch:
        # Get per-microbatch normalization factor
        weight_sum = loss_weights.sum()
        mask_sum = loss_mask.sum()
        weight_normalizer = torch.maximum(weight_sum, torch.ones_like(weight_sum))
        loss_weights = (mask_sum * loss_weights) / weight_normalizer

    # Apply loss weights and loss mask to the loss
    loss = loss * loss_weights * loss_mask

    return loss



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F

from nemo.core.classes.common import typecheck
from nemo.core.classes.exportable import Exportable
from nemo.core.classes.module import NeuralModule
from nemo.core.neural_types import EncodedRepresentation, LengthsType, NeuralType, SpectrogramType
from nemo.core.neural_types.elements import ProbsType

__all__ = ['MSDD_module']


class ConvLayer(nn.Module):
    def __init__(self, in_channels=1, out_channels=1, kernel_size=(3, 1), stride=(1, 1)):
        super(ConvLayer, self).__init__()
        self.cnn = nn.Sequential(
            nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride),
            nn.ReLU(),
            nn.BatchNorm2d(out_channels, eps=0.001, momentum=0.99),
        )

    def forward(self, feature):
        feature = self.cnn(feature)
        return feature


class MSDD_module(NeuralModule, Exportable):
    """
    Multi-scale Diarization Decoder (MSDD) for overlap-aware diarization and improved diarization accuracy from clustering diarizer.
    Based on the paper: Taejin Park et. al, "Multi-scale Speaker Diarization with Dynamic Scale Weighting", Interspeech 2022.
    Arxiv version: https://arxiv.org/pdf/2203.15974.pdf

    Args:
        num_spks (int):
            Max number of speakers that are processed by the model. In `MSDD_module`, `num_spks=2` for pairwise inference.
        hidden_size (int):
            Number of hidden units in sequence models and intermediate layers.
        num_lstm_layers (int):
            Number of the stacked LSTM layers.
        dropout_rate (float):
            Dropout rate for linear layers, CNN and LSTM.
        cnn_output_ch (int):
            Number of channels per each CNN layer.
        emb_dim (int):
            Dimension of the embedding vectors.
        scale_n (int):
            Number of scales in multi-scale system.
        clamp_max (float):
            Maximum value for limiting the scale weight values.
        conv_repeat (int):
            Number of CNN layers after the first CNN layer.
        weighting_scheme (str):
            Name of the methods for estimating the scale weights.
        context_vector_type (str):
            If 'cos_sim', cosine similarity values are used for the input of the sequence models.
            If 'elem_prod', element-wise product values are used for the input of the sequence models.
    """

    @property
    def output_types(self):
        """
        Return definitions of module output ports.
        """
        return OrderedDict(
            {
                "probs": NeuralType(('B', 'T', 'C'), ProbsType()),
                "scale_weights": NeuralType(('B', 'T', 'C', 'D'), ProbsType()),
            }
        )

    @property
    def input_types(self):
        """
        Return  definitions of module input ports.
        """
        return OrderedDict(
            {
                "ms_emb_seq": NeuralType(('B', 'T', 'C', 'D'), SpectrogramType()),
                "length": NeuralType(tuple('B'), LengthsType()),
                "ms_avg_embs": NeuralType(('B', 'C', 'D', 'C'), EncodedRepresentation()),
                "targets": NeuralType(('B', 'T', 'C'), ProbsType()),
            }
        )

    def init_weights(self, m):
        if type(m) == nn.Linear:
            torch.nn.init.xavier_uniform_(m.weight)
            m.bias.data.fill_(0.01)
        elif type(m) in [nn.GRU, nn.LSTM, nn.RNN]:
            for name, param in m.named_parameters():
                if 'weight_ih' in name:
                    torch.nn.init.xavier_uniform_(param.data)
                elif 'weight_hh' in name:
                    torch.nn.init.orthogonal_(param.data)
                elif 'bias' in name:
                    param.data.fill_(0.01)

    def __init__(
        self,
        num_spks: int = 2,
        hidden_size: int = 256,
        num_lstm_layers: int = 2,
        dropout_rate: float = 0.5,
        cnn_output_ch: int = 16,
        emb_dim: int = 192,
        scale_n: int = 5,
        clamp_max: float = 1.0,
        conv_repeat: int = 1,
        weighting_scheme: str = 'conv_scale_weight',
        context_vector_type: str = 'cos_sim',
    ):
        super().__init__()
        self._speaker_model = None
        self.batch_size: int = 1
        self.length: int = 50
        self.emb_dim: int = emb_dim
        self.num_spks: int = num_spks
        self.scale_n: int = scale_n
        self.cnn_output_ch: int = cnn_output_ch
        self.conv_repeat: int = conv_repeat
        self.chan: int = 2
        self.eps: float = 1e-6
        self.num_lstm_layers: int = num_lstm_layers
        self.weighting_scheme: str = weighting_scheme
        self.context_vector_type: bool = context_vector_type

        self.softmax = torch.nn.Softmax(dim=2)
        self.cos_dist = torch.nn.CosineSimilarity(dim=3, eps=self.eps)
        self.lstm = nn.LSTM(
            hidden_size,
            hidden_size,
            num_layers=self.num_lstm_layers,
            batch_first=True,
            bidirectional=True,
            dropout=dropout_rate,
        )

        if self.weighting_scheme == 'conv_scale_weight':
            self.conv = nn.ModuleList(
                [
                    ConvLayer(
                        in_channels=1,
                        out_channels=cnn_output_ch,
                        kernel_size=(self.scale_n + self.scale_n * num_spks, 1),
                        stride=(1, 1),
                    )
                ]
            )
            for conv_idx in range(1, conv_repeat + 1):
                self.conv.append(
                    ConvLayer(
                        in_channels=1, out_channels=cnn_output_ch, kernel_size=(self.cnn_output_ch, 1), stride=(1, 1)
                    )
                )
            self.conv_bn = nn.ModuleList()
            for conv_idx in range(self.conv_repeat + 1):
                self.conv_bn.append(nn.BatchNorm2d(self.emb_dim, affine=False))
            self.conv_to_linear = nn.Linear(emb_dim * cnn_output_ch, hidden_size)
            self.linear_to_weights = nn.Linear(hidden_size, self.scale_n)

        elif self.weighting_scheme == 'attn_scale_weight':
            self.W_a = nn.Linear(emb_dim, emb_dim, bias=False)
            nn.init.eye_(self.W_a.weight)
        else:
            raise ValueError(f"No such weighting scheme as {self.weighting_scheme}")

        self.hidden_to_spks = nn.Linear(2 * hidden_size, self.num_spks)
        if self.context_vector_type == "cos_sim":
            self.dist_to_emb = nn.Linear(self.scale_n * self.num_spks, hidden_size)
            self.dist_to_emb.apply(self.init_weights)
        elif self.context_vector_type == "elem_prod":
            self.product_to_emb = nn.Linear(self.emb_dim * self.num_spks, hidden_size)
        else:
            raise ValueError(f"No such context vector type as {self.context_vector_type}")

        self.dropout = nn.Dropout(dropout_rate)
        self.hidden_to_spks.apply(self.init_weights)
        self.lstm.apply(self.init_weights)
        self.clamp_max = clamp_max

    def core_model(self, ms_emb_seq, length, ms_avg_embs, targets):
        """
        Core model that accepts multi-scale cosine similarity values and estimates per-speaker binary label.

        Args:
            ms_emb_seq (Tensor):
                Multiscale input embedding sequence
                Shape: (batch_size, length, scale_n, emb_dim)
            length (Tensor):
                The actual length of embedding sequences without zero padding
                Shape: (batch_size,)
            ms_avg_embs (Tensor):
                Cluster-average speaker embedding vectors.
                Shape: (batch_size, scale_n, self.emb_dim, max_spks)
            targets (Tensor):
                Ground-truth labels for the finest segment.
                Shape: (batch_size, feats_len, max_spks)

        Returns:
            preds (Tensor):
                Predicted binary speaker label for each speaker.
                Shape: (batch_size, feats_len, max_spks)
            scale_weights (Tensor):
                Multiscale weights per each base-scale segment.
                Shape: (batch_size, length, scale_n, max_spks)

        """
        self.batch_size = ms_emb_seq.shape[0]
        self.length = ms_emb_seq.shape[1]
        self.emb_dim = ms_emb_seq.shape[-1]

        _ms_emb_seq = ms_emb_seq.unsqueeze(4).expand(-1, -1, -1, -1, self.num_spks)
        ms_emb_seq_single = ms_emb_seq
        ms_avg_embs = ms_avg_embs.unsqueeze(1).expand(-1, self.length, -1, -1, -1)

        ms_avg_embs_perm = ms_avg_embs.permute(0, 1, 2, 4, 3).reshape(self.batch_size, self.length, -1, self.emb_dim)

        if self.weighting_scheme == "conv_scale_weight":
            scale_weights = self.conv_scale_weights(ms_avg_embs_perm, ms_emb_seq_single)
        elif self.weighting_scheme == "attn_scale_weight":
            scale_weights = self.attention_scale_weights(ms_avg_embs_perm, ms_emb_seq_single)
        else:
            raise ValueError(f"No such weighting scheme as {self.weighting_scheme}")
        scale_weights = scale_weights.to(ms_emb_seq.device)

        if self.context_vector_type == "cos_sim":
            context_emb = self.cosine_similarity(scale_weights, ms_avg_embs, _ms_emb_seq)
        elif self.context_vector_type == "elem_prod":
            context_emb = self.element_wise_product(scale_weights, ms_avg_embs, _ms_emb_seq)
        else:
            raise ValueError(f"No such context vector type as {self.context_vector_type}")

        context_emb = self.dropout(F.relu(context_emb))
        lstm_output = self.lstm(context_emb)
        lstm_hidden_out = self.dropout(F.relu(lstm_output[0]))
        spk_preds = self.hidden_to_spks(lstm_hidden_out)
        preds = nn.Sigmoid()(spk_preds)
        return preds, scale_weights

    def element_wise_product(self, scale_weights, ms_avg_embs, ms_emb_seq):
        """
        Calculate element wise product values among cluster-average embedding vectors and input embedding vector sequences.
        This function is selected by assigning `self.context_vector_type = "elem_prod"`. `elem_prod` method usually takes more
        time to converge compared to `cos_sim` method.

        Args:
            scale_weights (Tensor):
                Multiscale weight vector.
                Shape: (batch_size, feats_len, scale_n, max_spks)
            ms_avg_embs_perm (Tensor):
                Tensor containing cluster-average speaker embeddings for each scale.
                Shape: (batch_size, length, scale_n, emb_dim)
            ms_emb_seq (Tensor):
                Tensor containing multi-scale speaker embedding sequences. `ms_emb_seq` is a single channel input from the
                given audio stream input.
                Shape: (batch_size, length, num_spks, emb_dim)

        Returns:
            context_emb (Tensor):
                Output of `dist_to_emb` linear layer containing context for speaker label estimation.
        """
        scale_weight_flatten = scale_weights.reshape(self.batch_size * self.length, self.num_spks, self.scale_n)
        ms_avg_embs_flatten = ms_avg_embs.reshape(
            self.batch_size * self.length, self.scale_n, self.emb_dim, self.num_spks
        )
        ms_emb_seq_flatten = ms_emb_seq.reshape(-1, self.scale_n, self.emb_dim)
        ms_emb_seq_flatten_rep = ms_emb_seq_flatten.unsqueeze(3).reshape(-1, self.scale_n, self.emb_dim, self.num_spks)
        elemwise_product = ms_avg_embs_flatten * ms_emb_seq_flatten_rep
        context_vectors = torch.bmm(
            scale_weight_flatten.reshape(self.batch_size * self.num_spks * self.length, 1, self.scale_n),
            elemwise_product.reshape(self.batch_size * self.num_spks * self.length, self.scale_n, self.emb_dim),
        )
        context_vectors = context_vectors.reshape(self.batch_size, self.length, self.emb_dim * self.num_spks)
        context_emb = self.product_to_emb(context_vectors)
        return context_emb

    def cosine_similarity(self, scale_weights, ms_avg_embs, _ms_emb_seq):
        """
        Calculate cosine similarity values among cluster-average embedding vectors and input embedding vector sequences.
        This function is selected by assigning self.context_vector_type = "cos_sim".

        Args:
            scale_weights (Tensor):
                Multiscale weight vector.
                Shape: (batch_size, feats_len, scale_n, max_spks)
            ms_avg_embs_perm (Tensor):
                Tensor containing cluster-average speaker embeddings for each scale.
                Shape: (batch_size, length, scale_n, emb_dim)
            _ms_emb_seq (Tensor):
                Tensor containing multi-scale speaker embedding sequences. `ms_emb_seq` is a single channel input from the
                given audio stream input.
                Shape: (batch_size, length, num_spks, emb_dim)

        Returns:
            context_emb (Tensor):
                Output of `dist_to_emb` linear layer containing context for speaker label estimation.
        """
        cos_dist_seq = self.cos_dist(_ms_emb_seq, ms_avg_embs)
        context_vectors = torch.mul(scale_weights, cos_dist_seq)
        context_vectors = context_vectors.view(self.batch_size, self.length, -1)
        context_emb = self.dist_to_emb(context_vectors)
        return context_emb

    def attention_scale_weights(self, ms_avg_embs_perm, ms_emb_seq):
        """
        Use weighted inner product for calculating each scale weight. W_a matrix has (emb_dim * emb_dim) learnable parameters
        and W_a matrix is initialized with an identity matrix. Compared to "conv_scale_weight" method, this method shows more evenly
        distributed scale weights.

        Args:
            ms_avg_embs_perm (Tensor):
                Tensor containing cluster-average speaker embeddings for each scale.
                Shape: (batch_size, length, scale_n, emb_dim)
            ms_emb_seq (Tensor):
                Tensor containing multi-scale speaker embedding sequences. `ms_emb_seq` is input from the
                given audio stream input.
                Shape: (batch_size, length, num_spks, emb_dim)

        Returns:
            scale_weights (Tensor):
                Weight vectors that determine the weight of each scale.
                Shape: (batch_size, length, num_spks, emb_dim)
        """
        self.W_a(ms_emb_seq.flatten(0, 1))
        mat_a = self.W_a(ms_emb_seq.flatten(0, 1))
        mat_b = ms_avg_embs_perm.flatten(0, 1).permute(0, 2, 1)

        weighted_corr = torch.matmul(mat_a, mat_b).reshape(-1, self.scale_n, self.scale_n, self.num_spks)
        scale_weights = torch.sigmoid(torch.diagonal(weighted_corr, dim1=1, dim2=2))
        scale_weights = scale_weights.reshape(self.batch_size, self.length, self.scale_n, self.num_spks)
        scale_weights = self.softmax(scale_weights)
        return scale_weights

    def conv_scale_weights(self, ms_avg_embs_perm, ms_emb_seq_single):
        """
        Use multiple Convnet layers to estimate the scale weights based on the cluster-average embedding and
        input embedding sequence.

        Args:
            ms_avg_embs_perm (Tensor):
                Tensor containing cluster-average speaker embeddings for each scale.
                Shape: (batch_size, length, scale_n, emb_dim)
            ms_emb_seq_single (Tensor):
                Tensor containing multi-scale speaker embedding sequences. ms_emb_seq_single is input from the
                given audio stream input.
                Shape: (batch_size, length, num_spks, emb_dim)

        Returns:
            scale_weights (Tensor):
                Weight vectors that determine the weight of each scale.
                Shape: (batch_size, length, num_spks, emb_dim)
        """
        ms_cnn_input_seq = torch.cat([ms_avg_embs_perm, ms_emb_seq_single], dim=2)
        ms_cnn_input_seq = ms_cnn_input_seq.unsqueeze(2).flatten(0, 1)

        conv_out = self.conv_forward(
            ms_cnn_input_seq, conv_module=self.conv[0], bn_module=self.conv_bn[0], first_layer=True
        )
        for conv_idx in range(1, self.conv_repeat + 1):
            conv_out = self.conv_forward(
                conv_input=conv_out,
                conv_module=self.conv[conv_idx],
                bn_module=self.conv_bn[conv_idx],
                first_layer=False,
            )

        lin_input_seq = conv_out.view(self.batch_size, self.length, self.cnn_output_ch * self.emb_dim)
        hidden_seq = self.conv_to_linear(lin_input_seq)
        hidden_seq = self.dropout(F.leaky_relu(hidden_seq))
        scale_weights = self.softmax(self.linear_to_weights(hidden_seq))
        scale_weights = scale_weights.unsqueeze(3).expand(-1, -1, -1, self.num_spks)
        return scale_weights

    def conv_forward(self, conv_input, conv_module, bn_module, first_layer=False):
        """
        A module for convolutional neural networks with 1-D filters. As a unit layer batch normalization, non-linear layer and dropout
        modules are included.

        Note:
            If `first_layer=True`, the input shape is set for processing embedding input.
            If `first_layer=False`, then the input shape is set for processing the output from another `conv_forward` module.

        Args:
            conv_input (Tensor):
                Reshaped tensor containing cluster-average embeddings and multi-scale embedding sequences.
                Shape: (batch_size*length, 1, scale_n*(num_spks+1), emb_dim)
            conv_module (ConvLayer):
                ConvLayer instance containing torch.nn.modules.conv modules.
            bn_module (torch.nn.modules.batchnorm.BatchNorm2d):
                Predefined Batchnorm module.
            first_layer (bool):
                Boolean for switching between the first layer and the others.
                Default: `False`

        Returns:
            conv_out (Tensor):
                Convnet output that can be fed to another ConvLayer module or linear layer.
                Shape: (batch_size*length, 1, cnn_output_ch, emb_dim)
        """
        conv_out = conv_module(conv_input)
        conv_out = conv_out.permute(0, 2, 1, 3) if not first_layer else conv_out
        conv_out = conv_out.reshape(self.batch_size, self.length, self.cnn_output_ch, self.emb_dim)
        conv_out = conv_out.unsqueeze(2).flatten(0, 1)
        conv_out = bn_module(conv_out.permute(0, 3, 2, 1)).permute(0, 3, 2, 1)
        conv_out = self.dropout(F.leaky_relu(conv_out))
        return conv_out

    @typecheck()
    def forward(self, ms_emb_seq, length, ms_avg_embs, targets):
        preds, scale_weights = self.core_model(ms_emb_seq, length, ms_avg_embs, targets)
        return preds, scale_weights

    def input_example(self):
        """
        Generate input examples for tracing etc.

        Returns (tuple):
            A tuple of input examples.
        """
        device = next(self.parameters()).device
        lens = torch.full(size=(input_example.shape[0],), fill_value=123, device=device)
        input_example = torch.randn(1, lens, self.scale_n, self.emb_dim, device=device)
        avg_embs = torch.randn(1, self.scale_n, self.emb_dim, self.num_spks, device=device)
        targets = torch.randn(1, lens, self.num_spks).round().float()
        return tuple([input_example, lens, avg_embs, targets])



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy

import numpy as np
import pytest
from lhotse import SupervisionSegment
from lhotse.testing.dummies import dummy_cut

from nemo.collections.common.data.lhotse.sampling import (
    DurationFilter,
    TokenCountFilter,
    TokenPerSecondFilter,
    TokenPerTokenFilter,
)
from nemo.collections.common.data.lhotse.text_adapters import NeMoSFTExample, SourceTargetTextExample, TextExample


@pytest.fixture
def cut():
    c = dummy_cut(0, duration=1.0, supervisions=[SupervisionSegment("", "", 0, 1.0, text="dummy")])
    c.supervisions[0].tokens = [1, 37, 12, 2]
    return c


def test_cut_duration_filter(cut):
    f = DurationFilter(0, 10)
    assert f(cut) == True

    f = DurationFilter(0, 0.5)
    assert f(cut) == False

    f = DurationFilter(1.5, 2.0)
    assert f(cut) == False


def test_cut_token_per_second_filter(cut):
    f = TokenPerSecondFilter(tps_min=0.0, tps_max=5.0)
    assert f(cut) == True

    f = TokenPerSecondFilter(tps_min=0.0, tps_max=1.0)
    assert f(cut) == False

    f = TokenPerSecondFilter(tps_min=10.0, tps_max=12.0)
    assert f(cut) == False


def test_cut_passes_by_token_count_and_tpt_filter(cut):
    assert TokenCountFilter(1, 10, measure_total_length=True)(cut) == True
    assert TokenPerTokenFilter(1, 10)(cut) == True


def test_cut_passes_by_token_count_and_tpt_filter(cut):
    assert TokenCountFilter(1, 10, measure_total_length=True)(cut) == True
    assert TokenPerTokenFilter(1, 10)(cut) == True


@pytest.fixture
def src_tgt_example():
    return SourceTargetTextExample(
        source=TextExample("", tokens=np.array([1, 37, 12, 2])),
        target=TextExample("", tokens=np.array([1, 1823, 1245, 2446, 1038, 2])),
    )


def test_src_tgt_token_filter_requires_prompt_formatting(src_tgt_example):
    with pytest.raises(RuntimeError):
        TokenCountFilter(0, 1, True)(src_tgt_example)


def test_src_tgt_passes_by_duration_filter(src_tgt_example):
    assert DurationFilter(1, 10)(src_tgt_example) == True
    assert TokenPerSecondFilter(1, 10)(src_tgt_example) == True


def test_src_tgt_token_filter(src_tgt_example):
    example = deepcopy(src_tgt_example)
    example.input_ids = np.concatenate((example.source.tokens, example.target.tokens))
    example.context_ids = example.source.tokens
    example.answer_ids = example.target.tokens

    """
    Input length measurement / encoder-decoder models / 2D bucketing
    """
    f = TokenCountFilter(1, 5, measure_total_length=False)
    assert f(example) == True

    f = TokenCountFilter(1, 3, measure_total_length=False)
    assert f(example) == False

    f = TokenCountFilter(10, 30, measure_total_length=False)
    assert f(example) == False

    """
    Total length measurement / decoder-only models / 1D bucketing
    """
    f = TokenCountFilter(1, 5, measure_total_length=True)
    assert f(example) == False

    f = TokenCountFilter(1, 20, measure_total_length=True)
    assert f(example) == True

    f = TokenCountFilter(1, 3, measure_total_length=True)
    assert f(example) == False

    f = TokenCountFilter(20, 30, measure_total_length=True)
    assert f(example) == False


@pytest.fixture
def nemo_sft_example():
    example = NeMoSFTExample(
        data={
            "system": "",
            "mask": "User",
            "dataset": "",
            "conversations": [
                {
                    "from": "User",
                    "value": "Hi, how are you?",
                },
                {
                    "from": "Assistant",
                    "value": "Good day, I'm a useful assistant.",
                },
            ],
        },
    )
    return example


def test_nemo_sft_token_filter_requires_prompt_formatting(nemo_sft_example):
    with pytest.raises(RuntimeError):
        TokenCountFilter(0, 1, True)(nemo_sft_example)


def test_nemo_sft_passes_by_duration_filter(nemo_sft_example):
    assert DurationFilter(1, 10)(nemo_sft_example) == True
    assert TokenPerSecondFilter(1, 10)(nemo_sft_example) == True


def test_nemo_sft_token_filter(nemo_sft_example):
    example = deepcopy(nemo_sft_example)
    example.input_ids = np.array([1, 123, 3425, 123, 2345, 324, 54, 2])
    example.context_ids = np.array([1, 123, 3425])
    example.answer_ids = np.array([123, 2345, 324, 54, 2])

    """
    Input length measurement / encoder-decoder models / 2D bucketing
    """
    f = TokenCountFilter(1, 5, measure_total_length=False)
    assert f(example) == True

    f = TokenCountFilter(1, 2, measure_total_length=False)
    assert f(example) == False

    f = TokenCountFilter(10, 30, measure_total_length=False)
    assert f(example) == False

    """
    Total length measurement / decoder-only models / 1D bucketing
    """
    f = TokenCountFilter(1, 5, measure_total_length=True)
    assert f(example) == False

    f = TokenCountFilter(1, 20, measure_total_length=True)
    assert f(example) == True

    f = TokenCountFilter(1, 3, measure_total_length=True)
    assert f(example) == False

    f = TokenCountFilter(10, 30, measure_total_length=True)
    assert f(example) == False



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
from concurrent.futures import ProcessPoolExecutor
from io import BytesIO
from multiprocessing import get_start_method
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Callable, Dict, Optional, Union

import torch
from lightning.fabric.plugins.io.checkpoint_io import CheckpointIO

from nemo.utils import logging
from nemo.utils.s3_utils import (
    DEFAULT_CHUNK_SIZE_MB,
    DEFAULT_MAX_READ_CONCURRENCY,
    DEFAULT_MAX_WRITE_CONCURRENCY,
    SHARED_MEM_DIR,
    S3Utils,
)


class S3CheckpointIO(CheckpointIO):
    """A custom S3CheckpointIO module that supports checkpoint reading/writing with s3 when filepath
    is a s3 url.
    """

    def __init__(
        self,
        dirpath: str,
        chunk_size_MB=DEFAULT_CHUNK_SIZE_MB,
        max_read_concurrency=DEFAULT_MAX_READ_CONCURRENCY,
        max_write_concurrency=DEFAULT_MAX_WRITE_CONCURRENCY,
        async_checkpointing=False,
    ):
        """
        Initialize the transfer configuration with custom values.

        This method overrides the default TransferConfig values in boto3.
        See https://boto3.amazonaws.com/v1/documentation/api/latest/_modules/boto3/s3/transfer.html#TransferConfig

        Args:
            chunk_size_MB (int, optional): The size of chunks to use when transferring files.
                Default is 64 (MB).
            max_read_concurrency (int, optional): The maximum number of threads that will be making
                requests to perform a download. Default is 15.
            max_write_concurrency (int, optional): The maximum number of threads that will be making
                requests to perform an upload. Default is 10.
            async_checkpointing (bool, optional): Uses a ProcessPoolExecutor to do the main saving logic.
                This feature should be used with save_top_k as it's possible a previous checkpoint is removed while
                the current checkpoint write fails.
        """
        if not S3Utils.is_s3_url(dirpath):
            raise AssertionError(
                f"Error attempting to initialize an S3CheckpointIO when {dirpath} is not an S3 url. Please use TorchCheckpointIO when using a non-S3 dirpath."
            )

        self.chunk_size_MB = chunk_size_MB
        self.max_read_concurrency = max_read_concurrency
        self.max_write_concurrency = max_write_concurrency
        self._async_checkpointing = async_checkpointing
        '''
        When using shared memory, we create a temporary file to hold the checkpoint before uploading to S3. 
        This list will track those temporary files, and clean up any leaked files that are still around during teardown. 
        '''
        self._temp_files = []

        if self.async_checkpointing:
            # create an executor that will asynchronously run functions
            self._executor = ProcessPoolExecutor(max_workers=1) if self.async_checkpointing else None

            # Eager creating a subprocess now so that forked subprocess does not inherit cuda context from parent
            if get_start_method() == 'fork' and torch.cuda.is_initialized() is True:
                raise Exception(
                    f'torch.cuda should not be initialized when checkpointing subprocess is created by fork method'
                )
            logging.info(f'Creating asynchronous checkpointing subprocess')
            future = self._executor.submit(dummy_func)
            try:
                future.result()
                logging.info(f'Asynchronous heckpointing subprocess created successfully')
            except Exception as e:
                logging.error(f'Failed to create asynchronous checkpointing subprocess, exception: {e}')
                raise e
            self._futures = []

        super().__init__()

    @property
    def async_checkpointing(self):
        return self._async_checkpointing

    def _serialize_checkpoint_to_shm(self, checkpoint: Dict, path: str) -> str:
        """
        Returns:
            filename of the temporary file in shared memory.
        """
        start_time = time.perf_counter()
        tempfile = NamedTemporaryFile(dir=SHARED_MEM_DIR, delete=False)
        torch.save(checkpoint, tempfile)
        logging.info(
            f'Time elapsed saving checkpoint dict to {tempfile.name} for {path}: {(time.perf_counter() - start_time):.2f} seconds, rank {torch.distributed.get_rank()}'
        )
        del checkpoint
        return tempfile.name

    def _serialize_checkpoint_to_bytes(self, checkpoint: Dict, path: str) -> BytesIO:
        """
        Returns:
            The bytestring of the checkpoint.
        """
        ss = time.perf_counter()
        bytes = BytesIO()
        torch.save(checkpoint, bytes)
        tt = time.perf_counter() - ss
        logging.info(
            f'Time elapsed saving checkpoint dict to bytes for {path}: {tt:.2f} seconds, rank {torch.distributed.get_rank()}'
        )
        del checkpoint
        return bytes

    def _check_uploading_results_so_far(self):
        """
        self._future is a list of tuples of form (future, destination path, source path)
        This function checks the result of all the futures, and updates the self._futures list appropriately.
        It also updates the list of self._temp_files, which is used to clean up leaked temporary files in SHARED_MEM during teardown.
        """
        if not self._futures:
            return
        start_time = time.perf_counter()
        done_futures = []
        in_progress_futures = []
        for item in self._futures:
            if item[0].done():
                done_futures.append(item)
            else:
                in_progress_futures.append(item)

        for item in done_futures:
            try:
                item[0].result()
            except Exception as e:
                logging.error(f'Failed to upload {item[2]} to {item[1]}, exception: {e}')
                raise e
            # If the future is complete, we can remove the temp file since we choose to clear the temp file when uploading.
            try:
                self._temp_files.remove(item[2])
            except:
                pass  # When not using shared memory, we do not append anything to the temp_files list, so remove will do nothing.
        self._futures = in_progress_futures
        logging.debug(
            f'Time elapsed checking uploading future results: {(time.perf_counter() - start_time):.2f} seconds'
        )

    def save_checkpoint(
        self, checkpoint: Dict[str, Any], path: Union[str, Path], storage_options: Optional[Any] = None
    ) -> None:
        # if we have a shared memory directory, we can serialize as a file to shared memory instead of as bytes.
        if os.path.exists(SHARED_MEM_DIR):
            localfile = self._serialize_checkpoint_to_shm(checkpoint, path)
            self._temp_files.append(localfile)
            saved_as_file = True
        else:
            bytes = self._serialize_checkpoint_to_bytes(checkpoint, path)
            saved_as_file = False

        if self.async_checkpointing:
            self._check_uploading_results_so_far()
            logging.info(f'Uploading checkpoint to {path} in asynchronous mode, rank {torch.distributed.get_rank()}')
            if saved_as_file:
                future = self._executor.submit(
                    _upload_file_to_s3, localfile, path, self.chunk_size_MB, self.max_write_concurrency, True
                )
                self._futures.append((future, path, localfile))
            else:
                future = self._executor.submit(
                    _upload_bytes_to_s3, bytes, path, self.chunk_size_MB, self.max_write_concurrency
                )
                self._futures.append((future, path, 'bytes'))
        else:
            logging.info(f'Uploading checkpoint to {path} in synchronous mode, rank {torch.distributed.get_rank()}')
            if saved_as_file:
                _upload_file_to_s3(localfile, path, self.chunk_size_MB, self.max_write_concurrency, True)
                self._temp_files.remove(localfile)
            else:
                _upload_bytes_to_s3(bytes, path, self.chunk_size_MB, self.max_write_concurrency)

    def load_checkpoint(
        self, path: Union[str, Path], map_location: Optional[Callable] = lambda storage, loc: storage
    ) -> Dict[str, Any]:
        if os.path.exists(SHARED_MEM_DIR):
            with NamedTemporaryFile(dir=SHARED_MEM_DIR, delete=True) as tempfile:
                logging.info(
                    f'Loading checkpoint {path} into a temp file in shared memory {tempfile.name}, rank {torch.distributed.get_rank()}'
                )
                S3Utils.download_s3_file_to_path(
                    s3_path=path,
                    file_path=tempfile.name,
                    chunk_size_MB=self.chunk_size_MB,
                    max_concurrency=self.max_read_concurrency,
                )
                checkpoint = torch.load(tempfile.name)
        else:
            file_stream: BytesIO = S3Utils.download_s3_file_to_stream(
                s3_path=path, chunk_size_MB=self.chunk_size_MB, max_concurrency=self.max_read_concurrency
            )
            checkpoint = torch.load(file_stream)
        return checkpoint

    def remove_checkpoint(self, path: Union[str, Path]) -> None:
        if S3Utils.is_s3_url(path):
            S3Utils.remove_object(path)
        else:
            super().remove_checkpoint(path)

    def teardown(self) -> None:
        # this ensure we wait for final checkpoint to finish uploading at train end.
        rank = torch.distributed.get_rank()
        if self.async_checkpointing:
            logging.info(f'Entering teardown, waiting for all jobs to finish, rank {rank}')
            start_time = time.perf_counter()
            self._executor.shutdown(wait=True)
            logging.info(f'executor shut down after {(time.perf_counter() - start_time):.2f} seconds, rank {rank}')

        '''
        this will be non-empty at the end of training if using asynchronous uploading since the futures are not processed with _check_uploading_results_so_far.
        therefore, we check that the path exists first before trying to delete. 
        '''
        if self._temp_files:
            for tfile in self._temp_files:
                if os.path.exists(tfile):
                    try:
                        os.remove(tfile)
                    except Exception as e:
                        logging.info(f"Error occurred while deleting file {tfile}: {e}")


def _clean_up_conflicting_checkpoint(filepath: str) -> None:
    '''
    before saving to s3, clean up any existing object with the same prefix megatron_gpt+step_count
    e.g. before we save "megatron_gpt--step=1400-validation_loss=6.32-consumed_samples=55920.0-last.ckpt"
    we need to clean up "megatron_gpt--step=1400-validation_loss=xxx-consumed_samples=yyy-last.ckpt"
    so that in case later we need to resume from step 1400, it has a single checkpoint file at step 1400
    '''

    if S3Utils.is_s3_url(filepath):
        prefix_with_step = S3Utils.parse_prefix_with_step(filepath)
        logging.info(f'Looking for conflicting checkpoint under prefix {prefix_with_step}')

        conflict_last_ckpts = S3Utils.find_files_with_suffix(
            base_path=prefix_with_step, suffix='last.ckpt', return_key_only=False
        )
        for last_ckpt in conflict_last_ckpts:
            logging.info(f'Cleaning up conflicting last ckpt {last_ckpt} before saving {filepath}')
            S3Utils.remove_object(last_ckpt)


def _upload_file_to_s3(localfile, path, chunk_size_MB, max_write_concurrency, remove_file):
    try:
        _clean_up_conflicting_checkpoint(path)
        S3Utils.upload_file(localfile, path, chunk_size_MB, max_write_concurrency, remove_file)
    except Exception as e:
        raise e


def _upload_bytes_to_s3(bytes, path, chunk_size_MB, max_write_concurrency):
    try:
        _clean_up_conflicting_checkpoint(path)
        S3Utils.upload_file_stream_to_s3(bytes, path, chunk_size_MB, max_write_concurrency)
    except Exception as e:
        raise e


def dummy_func():
    time.sleep(0.01)



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import pytest
import torch

from nemo.collections.asr.parts.utils.asr_confidence_utils import (
    get_confidence_aggregation_bank,
    get_confidence_measure_bank,
)

# Initialize probability vectors
VOCAB_SIZES = (100, 1000, 10000)
ONE_VEC_SET, ZERO_VEC_SET, RAND_VEC_SET, OVERFIT_RAND_VEC_SET = {}, {}, {}, {}
for vocab_size in VOCAB_SIZES:
    # batch size 2 to test different positions of probability one
    ONE_VEC_SET[vocab_size] = torch.nan_to_num(
        torch.cat(
            [
                torch.tensor([[0] + [float('-inf')] * (vocab_size - 1)]),
                torch.tensor([[float('-inf')] * (vocab_size - 3) + [0] + [float('-inf')] * 2]),
            ]
        )
    )
    ZERO_VEC_SET[vocab_size] = torch.nan_to_num(torch.tensor([[math.log(1 / vocab_size)] * vocab_size] * 2))
    # batch size 1
    rand_logit = torch.rand((1, vocab_size))
    rand_logit_overfit = rand_logit.clone()
    rand_logit_overfit[0, 0] += vocab_size
    RAND_VEC_SET[vocab_size] = torch.nan_to_num(torch.nn.functional.log_softmax(rand_logit, -1))
    OVERFIT_RAND_VEC_SET[vocab_size] = torch.nan_to_num(torch.nn.functional.log_softmax(rand_logit_overfit, -1))
AGGREGATION_VEC_SIMPLE = [0.0, 0.5, 1]

TOL_DEGREE = 6
TOL = 1 / math.pow(10, TOL_DEGREE)


def get_measure_parametrize_ranges():
    confidence_measure_bank = {}
    alpha_range = (0.25, 0.5, 1.0)
    bank_exception = None
    try:
        confidence_measure_bank = get_confidence_measure_bank()
    except Exception as e:
        alpha_range = ()
        bank_exception = e
    return confidence_measure_bank, alpha_range, bank_exception


def get_aggregation_parametrize_ranges():
    confidence_aggregation_bank = {}
    bank_exception = None
    try:
        confidence_aggregation_bank = get_confidence_aggregation_bank()
    except Exception as e:
        bank_exception = e
    return confidence_aggregation_bank, bank_exception


class TestConfidenceMeasureBank:
    measure_bank, alphas, bank_build_exception = get_measure_parametrize_ranges()

    @pytest.mark.unit
    def test_measure_bank(self):
        if self.bank_build_exception is not None:
            raise self.bank_build_exception

        assert isinstance(self.measure_bank, dict)
        assert len(self.measure_bank) > 0

    @pytest.mark.unit
    @pytest.mark.parametrize('measure_name', measure_bank.keys())
    @pytest.mark.parametrize('alpha', alphas)
    @pytest.mark.parametrize('vocab_size', VOCAB_SIZES)
    def test_confidence_measures_one(self, measure_name, alpha, vocab_size):
        measure = self.measure_bank[measure_name]

        assert torch.allclose(measure(ONE_VEC_SET[vocab_size], vocab_size, alpha), torch.tensor([1.0, 1.0]), atol=TOL)

    @pytest.mark.unit
    @pytest.mark.parametrize('measure_name', measure_bank.keys())
    @pytest.mark.parametrize('alpha', alphas)
    @pytest.mark.parametrize('vocab_size', VOCAB_SIZES)
    def test_confidence_measures_zero(self, measure_name, alpha, vocab_size):
        measure = self.measure_bank[measure_name]

        assert torch.allclose(measure(ZERO_VEC_SET[vocab_size], vocab_size, alpha), torch.tensor([0.0, 0.0]), atol=TOL)

    @pytest.mark.unit
    @pytest.mark.parametrize('measure_name', measure_bank.keys())
    @pytest.mark.parametrize('alpha', alphas)
    @pytest.mark.parametrize('vocab_size', VOCAB_SIZES)
    def test_confidence_measures_partial_order(self, measure_name, alpha, vocab_size):
        measure = self.measure_bank[measure_name]
        value_normal = round(float(measure(RAND_VEC_SET[vocab_size], vocab_size, alpha)[0]), TOL_DEGREE)
        value_overfit = round(float(measure(OVERFIT_RAND_VEC_SET[vocab_size], vocab_size, alpha)[0]), TOL_DEGREE)

        assert 0 <= value_normal < value_overfit <= 1, (
            measure(RAND_VEC_SET[vocab_size], vocab_size, alpha),
            measure(OVERFIT_RAND_VEC_SET[vocab_size], vocab_size, alpha),
        )


class TestConfidenceAggregationBank:
    aggregation_bank, bank_build_exception = get_aggregation_parametrize_ranges()

    @pytest.mark.unit
    def test_aggregation_bank(self):
        if self.bank_build_exception is not None:
            raise self.bank_build_exception

        assert isinstance(self.aggregation_bank, dict)
        assert len(self.aggregation_bank) > 0

    @pytest.mark.unit
    @pytest.mark.parametrize('aggregation_name', aggregation_bank.keys())
    def test_confidence_agregation_simple(self, aggregation_name):
        # alaptev: would skipif work with parametrize arguments?
        if aggregation_name not in ("mean", "min", "max", "prod"):
            pytest.skip(f"{aggregation_name} is not a simple aggregation")
        aggregation = self.aggregation_bank[aggregation_name]
        if aggregation_name == "mean":
            assert aggregation(AGGREGATION_VEC_SIMPLE) == 0.5
        elif aggregation_name == "min":
            assert aggregation(AGGREGATION_VEC_SIMPLE) == 0.0
        if aggregation_name == "max":
            assert aggregation(AGGREGATION_VEC_SIMPLE) == 1.0
        if aggregation_name == "prod":
            assert aggregation(AGGREGATION_VEC_SIMPLE) == 0.0



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import json
from pathlib import Path

from prettytable import PrettyTable

import nemo.lightning as nl
from nemo.collections import llm, vlm
from nemo.collections.llm.inference.base import _setup_trainer_and_restore_model


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("--hf-path", type=str, default="models/llama_31_toy")
    parser.add_argument('--collection', type=str, default='llm')
    parser.add_argument("--model-type", type=str, help="Name of the model", default="LlamaModel")
    parser.add_argument("--model-config", type=str, help="Model config", default="Llama31Config8B")
    parser.add_argument("--output-path", type=str, help="Output NeMo2 model")
    parser.add_argument("--ignore-keys", nargs='+', type=str, default=[], help="Ignore keys")
    parser.add_argument("--trust-remote-code", action="store_true", help="if trust_remote_code")
    return parser


def count_parameters(model, ignore_keys=None):
    table = PrettyTable(["NeMo Modules", "Parameters"])
    total_params = 0
    for name, parameter in model.named_parameters():
        if not parameter.requires_grad:
            continue
        if ignore_keys is not None and name in ignore_keys:
            continue
        params = parameter.numel()
        table.add_row([name, params])
        total_params += params
    print(table)
    print(f"NeMo Total Trainable Params: {total_params}")
    return total_params


if __name__ == '__main__':
    args = get_parser().parse_args()

    if args.collection == 'llm':
        collection = llm
    elif args.collection == 'vlm':
        collection = vlm
    else:
        raise ValueError(f'Unrecognized collection {args.collection}')
    config = getattr(collection, args.model_config)()
    model = getattr(collection, args.model_type)(config=config)
    try:
        output_path = llm.import_ckpt(
            model=model,
            source=f"hf://{args.hf_path}",
            output_path=args.output_path,
            overwrite=True,
            trust_remote_code=args.trust_remote_code,
        )
    except TypeError:
        output_path = llm.import_ckpt(
            model=model,
            source=f"hf://{args.hf_path}",
            output_path=args.output_path,
            overwrite=True,
        )

    trainer = nl.Trainer(
        accelerator="cpu",
        devices=1,
        num_nodes=1,
        strategy=nl.MegatronStrategy(ckpt_load_strictness=False),
    )
    path = Path(output_path)
    model_nemo: nl.io.TrainerContext = nl.io.load_context(
        path=nl.ckpt_utils.ckpt_to_context_subdir(path), subpath="model"
    )
    _setup_trainer_and_restore_model(path=path, trainer=trainer, model=model_nemo)

    # Load HF Stats
    with open(f'{args.hf_path}/stats.json') as f:
        hf_stats = json.load(f)

    table = PrettyTable(["HuggingFace Modules", "Parameters"])
    for key, value in hf_stats.items():
        table.add_row([key, value])
    print(table)

    nemo_param_cnt = count_parameters(model_nemo, args.ignore_keys)
    hf_param_cnt = hf_stats['total_params']

    assert (
        hf_param_cnt == nemo_param_cnt
    ), f'Total converted params count does not match: NeMo model has {nemo_param_cnt} and HF model has {hf_param_cnt}.'
    print('HuggingFace -> NeMo conversion completed successfully.')



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# flake8: noqa
from nemo.collections.common.prompts.canary import CanaryPromptFormatter
from nemo.collections.common.prompts.canary2 import Canary2PromptFormatter
from nemo.collections.common.prompts.formatter import PromptFormatter
from nemo.collections.common.prompts.gemma import GemmaPromptFormatter
from nemo.collections.common.prompts.llama import Llama2PromptFormatter, Llama3PromptFormatter
from nemo.collections.common.prompts.mistral import MistralPromptFormatter
from nemo.collections.common.prompts.nemotron_h import NemotronHPromptFormatter
from nemo.collections.common.prompts.phi2 import (
    Phi2ChatPromptFormatter,
    Phi2CodePromptFormatter,
    Phi2QAPromptFormatter,
)
from nemo.collections.common.prompts.plain import PlainPromptFormatter
from nemo.collections.common.prompts.qwen import QwenPromptFormatter
from nemo.collections.common.prompts.t5nmt import T5NMTPromptFormatter



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings


def get_pad_id(tokenizer) -> int:
    pad_id = tokenizer.pad
    if pad_id is not None:
        return pad_id
    pad_id = tokenizer.unk_id
    if pad_id is not None:
        return pad_id
    warnings.warn(
        "The text tokenizer has no <pad> or <unk> tokens available, using ID 0 for padding (this may lead to silent bugs)."
    )
    return 0



# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import re
import unicodedata
from builtins import str as unicode
from typing import List, Tuple

__all__ = [
    "french_text_preprocessing",
    "chinese_text_preprocessing",
    "english_text_preprocessing",
    "any_locale_text_preprocessing",
    "spanish_text_preprocessing",
    "vietnamese_text_preprocessing",
    "italian_text_preprocessing",
    "any_locale_word_tokenize",
    "english_word_tokenize",
    "LATIN_CHARS_ALL",
    "normalize_unicode_text",
    "japanese_text_preprocessing",
]

# Derived from LJSpeech
_synoglyphs = {
    "'": ['’'],
    '"': ['”', '“'],
}
SYNOGLYPH2ASCII = {g: asc for asc, glyphs in _synoglyphs.items() for g in glyphs}

# Example of parsing by groups via _WORDS_RE_EN.
# Regular expression pattern groups:
#   1st group -- valid english words,
#   2nd group -- any substring starts from | to | (mustn't be nested), useful when you want to leave sequence unchanged,
#   3rd group -- punctuation marks or whitespaces.
# Text (first line) and mask of groups for every char (second line).
# config file must contain |EY1 EY1|, B, C, D, E, F, and G.

# define char set based on https://en.wikipedia.org/wiki/List_of_Unicode_characters
LATIN_ALPHABET_BASIC = "A-Za-z"
ACCENTED_CHARS = "À-ÖØ-öø-ÿ"
LATIN_CHARS_ALL = f"{LATIN_ALPHABET_BASIC}{ACCENTED_CHARS}"
_WORDS_RE_EN = re.compile(
    fr"([{LATIN_ALPHABET_BASIC}]+(?:[{LATIN_ALPHABET_BASIC}\-']*[{LATIN_ALPHABET_BASIC}]+)*)|(\|[^|]*\|)|([^{LATIN_ALPHABET_BASIC}|]+)"
)
_WORDS_RE_ANY_LOCALE = re.compile(
    fr"([{LATIN_CHARS_ALL}]+(?:[{LATIN_CHARS_ALL}\-']*[{LATIN_CHARS_ALL}]+)*)|(\|[^|]*\|)|([^{LATIN_CHARS_ALL}|]+)"
)


def english_text_preprocessing(text, lower=True):
    text = unicode(text)
    text = ''.join(char for char in unicodedata.normalize('NFD', text) if unicodedata.category(char) != 'Mn')
    text = ''.join(char if char not in SYNOGLYPH2ASCII else SYNOGLYPH2ASCII[char] for char in text)

    if lower:
        text = text.lower()

    return text


def any_locale_text_preprocessing(text: str) -> str:
    """
    Normalize unicode text with "NFC", and convert right single quotation mark (U+2019, decimal 8217) as an apostrophe.

    Args:
        text (str): the original input sentence.

    Returns: normalized text (str).
    """
    res = []
    for c in normalize_unicode_text(text):
        if c in ['’']:  # right single quotation mark (U+2019, decimal 8217) as an apostrophe
            res.append("'")
        else:
            res.append(c)

    return ''.join(res)


def normalize_unicode_text(text: str) -> str:
    """
    TODO @xueyang: Apply NFC form may be too aggressive since it would ignore some accented characters that do not exist
      in predefined German alphabet (nemo.collections.common.tokenizers.text_to_speech.ipa_lexicon.IPA_CHARACTER_SETS),
      such as 'é'. This is not expected. A better solution is to add an extra normalization with NFD to discard the
      diacritics and consider 'é' and 'e' produce similar pronunciations.

    Note that the tokenizer needs to run `unicodedata.normalize("NFC", x)` before calling `encode` function,
    especially for the characters that have diacritics, such as 'ö' in the German alphabet. 'ö' can be encoded as
    b'\xc3\xb6' (one char) as well as b'o\xcc\x88' (two chars). Without the normalization of composing two chars
    together and without a complete predefined set of diacritics, when the tokenizer reads the input sentence
    char-by-char, it would skip the combining diaeresis b'\xcc\x88', resulting in indistinguishable pronunciations
    for 'ö' and 'o'.

    Args:
        text (str): the original input sentence.

    Returns:
        NFC normalized sentence (str).
    """
    # normalize word with NFC form
    if not unicodedata.is_normalized("NFC", text):
        text = unicodedata.normalize("NFC", text)

    return text


def _word_tokenize(words: List[Tuple[str, str, str]], is_lower: bool = False) -> List[Tuple[List[str], bool]]:
    """
    Process a list of words and attach indicators showing if each word is unchangeable or not. Each word representation
    can be one of valid word, any substring starting from | to | (unchangeable word), or punctuation marks including
    whitespaces. This function will split unchanged strings by whitespaces and return them as `List[str]`. For example,

    .. code-block:: python
        [
            ('Hello', '', ''),  # valid word
            ('', '', ' '),  # punctuation mark
            ('World', '', ''),  # valid word
            ('', '', ' '),  # punctuation mark
            ('', '|NVIDIA unchanged|', ''),  # unchangeable word
            ('', '', '!')  # punctuation mark
        ]

    will be converted into,

    .. code-block:: python
        [
            (["Hello"], False),
            ([" "], False),
            (["World"], False),
            ([" "], False),
            (["NVIDIA", "unchanged"], True),
            (["!"], False)
        ]

    Args:
        words (List[str]): a list of tuples like `(maybe_word, maybe_without_changes, maybe_punct)` where each element
            corresponds to a non-overlapping match of either `_WORDS_RE_EN` or `_WORDS_RE_ANY_LOCALE`.
        is_lower (bool): a flag to trigger lowercase all words. By default, it is False.

    Returns: List[Tuple[List[str], bool]], a list of tuples like `(a list of words, is_unchanged)`.

    """
    result = []
    for word in words:
        maybe_word, maybe_without_changes, maybe_punct = word

        without_changes = False
        if maybe_word != '':
            if is_lower:
                token = [maybe_word.lower()]
            else:
                token = [maybe_word]
        elif maybe_punct != '':
            token = [maybe_punct]
        elif maybe_without_changes != '':
            without_changes = True
            token = maybe_without_changes[1:-1].split(" ")
        else:
            raise ValueError(
                f"This is not expected. Found empty string: <{word}>. "
                f"Please validate your regular expression pattern '_WORDS_RE_EN' or '_WORDS_RE_ANY_LOCALE'."
            )

        result.append((token, without_changes))

    return result


def english_word_tokenize(text: str) -> List[Tuple[List[str], bool]]:
    words = _WORDS_RE_EN.findall(text)
    return _word_tokenize(words, is_lower=True)


def any_locale_word_tokenize(text: str) -> List[Tuple[List[str], bool]]:
    words = _WORDS_RE_ANY_LOCALE.findall(text)
    return _word_tokenize(words)


def spanish_text_preprocessing(text: str) -> str:
    return text.lower()


def italian_text_preprocessing(text: str) -> str:
    return text.lower()


def chinese_text_preprocessing(text: str) -> str:
    return text


def french_text_preprocessing(text: str) -> str:
    return text.lower()


def vietnamese_text_preprocessing(text: str) -> str:
    return text.lower()


def japanese_text_preprocessing(text: str) -> str:
    return text.lower()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import List, Optional

import lightning.pytorch as L
from lightning.pytorch.utilities.types import OptimizerLRScheduler
from torch.optim import Optimizer

from nemo.lightning.io.mixin import IOMixin
from nemo.lightning.megatron_parallel import CallbackMethods


class LRSchedulerModule(L.Callback, CallbackMethods, IOMixin, ABC):
    """A module to standardize the learning rate scheduler setup and configuration.

    This class decouples the learning rate scheduler from the model, similar to how the LightningDataModule
    decouples data handling. It also acts as a Callback to hook into the training loop, which can be useful
    for adding custom all-reduces, logging, early stopping, etc. Next to that standard Lightning callback-event,
    this also supports hooking into the Megatron forward-backward function at a granular level.

    Example::

        class MyLRSchedulerModule(LRSchedulerModule):
            def setup(self, model, optimizer):
                # Custom setup logic
                ...

            def scheduler(self, model, optimizers):
                # Define and return the learning rate scheduler
                ...

    Methods:
        setup(model, optimizer): Sets up the learning rate scheduler.
        scheduler(model, optimizers): Abstract method to define the learning rate scheduler.
        __call__(model, optimizers): Calls the setup and scheduler methods.
    """

    def connect(self, model, optimizer) -> None:
        """Sets up the learning rate scheduler.

        Args:
            model: The model for which the scheduler is being set up.
            optimizer: The optimizer for which the scheduler is being set up.
        """
        ...

    @abstractmethod
    def scheduler(self, model, optimizers) -> OptimizerLRScheduler:
        """Abstract method to define the learning rate scheduler.

        Args:
            model: The model for which the scheduler is being defined.
            optimizers: The optimizers for which the scheduler is being defined.

        Returns:
            OptimizerLRScheduler: The learning rate scheduler.
        """
        raise NotImplementedError("The scheduler method should be implemented by subclasses.")

    def __call__(self, model, optimizers):
        """Calls the setup and scheduler methods.

        Args:
            model: The model for which the scheduler is being called.
            optimizers: The optimizers for which the scheduler is being called.

        Returns:
            OptimizerLRScheduler: The learning rate scheduler.
        """

        self.connect(model, optimizers)

        self._scheduler = self.scheduler(model, optimizers)

        if not isinstance(self._scheduler, (dict, tuple)):
            return optimizers, self._scheduler

        return self._scheduler


class OptimizerModule(L.Callback, CallbackMethods, IOMixin, ABC):
    """A module to standardize the optimizer setup and configuration.

    This class decouples the optimizer from the model, similar to how the LightningDataModule
    decouples data handling. It also acts as a Callback to hook into the training loop, which can be useful
    for adding custom all-reduces, logging, early stopping, etc. Next to that standard Lightning callback-event,
    this also supports hooking into the Megatron forward-backward function at a granular level.

    Attributes:
        lr_scheduler (Optional[LRSchedulerModule]): The learning rate scheduler module.

    Example::

        class MyOptimizerModule(OptimizerModule):
            def __init__(self, lr_scheduler=None):
                super().__init__(lr_scheduler)

            def setup(self, model):
                # Custom setup logic
                ...

            def optimizers(self, model):
                # Define and return the optimizers
                ...

    Methods:
        connect(model, trainer): Connects the optimizer module to the model and trainer.
        setup(model): Sets up the optimizer.
        optimizers(model): Abstract method to define the optimizers.
        __call__(model, megatron_parallel): Calls the setup and optimizers methods.
    """

    def __init__(self, lr_scheduler: Optional[LRSchedulerModule]):
        """Initializes the OptimizerModule.

        Args:
            lr_scheduler (Optional[LRSchedulerModule]): The learning rate scheduler module.
        """
        self.lr_scheduler = lr_scheduler

    def connect(self, model: L.LightningModule) -> None:
        """Connects the optimizer module to the model and trainer.

        Args:
            model (L.LightningModule): The model to which the optimizer module is being connected.
        """

        def custom_configure_optimizers(lightning_module_self, megatron_parallel=None):
            opt = self(lightning_module_self, megatron_parallel=megatron_parallel)
            return opt

        model.configure_optimizers = types.MethodType(custom_configure_optimizers, model)
        model.optim = self

        if hasattr(self, "__io__") and hasattr(model, "__io__"):
            if hasattr(model.__io__, "optim"):
                model.__io__.optim = deepcopy(self.__io__)

    @abstractmethod
    def optimizers(self, model) -> List[Optimizer]:
        """Abstract method to define the optimizers.

        Args:
            model: The model for which the optimizers are being defined.

        Returns:
            List[Optimizer]: The list of optimizers.
        """
        raise NotImplementedError("The optimizers method should be implemented by subclasses.")

    def on_train_batch_start(self, trainer, pl_module, batch, batch_idx) -> None:
        # pylint: disable=C0116
        if self._optimizers is not None:
            if len(self._optimizers[0].param_groups) > 0:
                lr = self._optimizers[0].param_groups[0]['lr']
            else:
                lr = 0.0
            pl_module.log('lr', lr, batch_size=1, prog_bar=True)

    def __call__(self, model: L.LightningModule, megatron_parallel=None) -> OptimizerLRScheduler:
        """Calls the setup and optimizers methods.

        Args:
            model (L.LightningModule): The model for which the optimizers are being called.
            megatron_parallel: Optional parallel model.

        Returns:
            OptimizerLRScheduler: The optimizers and optionally the learning rate scheduler.
        """
        _model = model if megatron_parallel is None else megatron_parallel
        callbacks = _model.trainer.callbacks
        if self not in callbacks:
            callbacks.append(self)
        if self.lr_scheduler is not None and self.lr_scheduler not in callbacks:
            callbacks.append(self.lr_scheduler)
        self._optimizers = self.optimizers(_model)

        _opt = self._optimizers[0] if len(self._optimizers) == 1 else self._optimizers

        if self.lr_scheduler is not None:
            with_scheduler = self.lr_scheduler(_model, _opt)

            return with_scheduler

        return self._optimizers



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import nemo_run as run

from nemo.lightning.pytorch.optim import PytorchOptimizerModule


@run.cli.factory
def pytorch_adam_with_flat_lr(
    lr: float = 1e-5,
) -> run.Config[PytorchOptimizerModule]:
    from torch.optim import Adam

    return run.Config(
        PytorchOptimizerModule,
        optimizer_fn=run.Partial(
            Adam,
            lr=lr,
            weight_decay=0.1,
            betas=(0.9, 0.95),
            eps=1e-8,
        ),
    )
